Require Import ZArith.
Require Import Bool.
Require Import List.

Definition nat_of_N (n:N) :=
  match n with
  | N0 => O
  | Npos p => nat_of_P p
  end.

Definition list_to_lambda (t:Set) (l:list t) (e0:t) (i:N) :=
  List.nth (nat_of_N i) l e0.

Definition bool_not := negb.
Definition bool_or := orb.
Definition bool_and := andb.

Definition Z_opp := Zopp. (* not used *)
Definition Z_plus := Zplus.
Definition Z_minus := Zminus.
Definition Z_times := Zmult.
Definition Z_div := Zdiv.
Definition Z_mod := Zmod.

Definition Z_eq := Zeq_bool.
Definition Z_ne := Zneq_bool.
Definition Z_lt := Zlt_bool.
Definition Z_le := Zle_bool.
Definition Z_gt := Zgt_bool.
Definition Z_ge := Zge_bool.

Definition N_plus := Nplus.

Definition N_minus (x y:N) : N :=
  match x, y with
  | xx, N0 => xx
  | N0, yy => N0
  | Npos xp, Npos yp =>
    match Zminus (Zpos xp) (Zpos yp) with
    | Zpos zp => Npos zp
    | _ => N0
    end
  end.

Definition N_times := Nmult.

Definition N_div (x y:N) : N :=
  match x, y with
  | xx, N0 => N0
  | N0, yy => N0
  | Npos xp, Npos yp =>
    match Zdiv (Zpos xp) (Zpos yp) with
    | Zpos zp => Npos zp
    | _ => N0
    end
  end.

Definition N_mod (x y:N) : N :=
  match x, y with
  | xx, N0 => N0
  | N0, yy => N0
  | Npos xp, Npos yp =>
    match Zmod (Zpos xp) (Zpos yp) with
    | Zpos zp => Npos zp
    | _ => N0
    end
  end.

Definition N_eq (x y:N) :=
  match (x ?= y)%N with
  | Eq => true
  | _ => false
  end.

Definition N_ne (x y:N) :=
  match (x ?= y)%N with
  | Eq => false
  | _ => true
  end.

Definition N_lt (x y:N) :=
  match (x ?= y)%N with
  | Lt => true
  | _ => false
  end.

Definition N_le (x y:N) :=
  match (x ?= y)%N with
  | Gt => false
  | _ => true
  end.

Definition N_gt (x y:N) :=
  match (x ?= y)%N with
  | Gt => true
  | _ => false
  end.

Definition N_ge (x y:N) :=
  match (x ?= y)%N with
  | Lt => false
  | _ => true
  end.

Definition N_lt_prop (x y:N) :=
  match (x ?= y)%N with
  | Lt => True
  | _ => False
  end.

(*
   Omega Definition
*)
Inductive Omega : Set :=
  | snat  : N -> Omega
  | sint  : Z -> Omega
  | sbool : bool -> Omega
  | arrow : Omega -> Omega -> Omega
  | tuple : N -> (N -> Omega) -> Omega
  | forallT : forall z:Type, (z -> Omega) -> Omega
  | existsT : forall z:Type, (z -> Omega) -> Omega
.

Notation "'Otuple' n [ x , .. , y ]" :=
  (tuple n (list_to_lambda Omega (cons x .. (cons y nil) ..) _)).
Notation "'Oemptytuple'" := (tuple 0 (list_to_lambda Omega nil _)).
Notation "'Oforall' x : a . b" :=
  (forallT a (fun x:a => b)) (at level 30, right associativity).
Notation "'Oexists' x : a . b" :=
  (existsT a (fun x:a => b)) (at level 30, right associativity).

(*
   OmegaK Definition
*)
Inductive OmegaK : Set :=
  | snatK  : N -> OmegaK
  | sintK  : Z -> OmegaK
  | sboolK : bool -> OmegaK
  | arrowK : OmegaK -> OmegaK
  | tupleK : N -> (N -> OmegaK) -> OmegaK
  | forallTK : forall z:Type, (z -> OmegaK) -> OmegaK
  | existsTK : forall z:Type, (z -> OmegaK) -> OmegaK
  | funcK : OmegaK -> OmegaK
.

Notation "'OKtuple' n [ x , .. , y ]" :=
  (tupleK n (list_to_lambda OmegaK (cons x .. (cons y nil) ..) _)).
Notation "'OKemptytuple'" := (tupleK 0 (list_to_lambda OmegaK nil _)).
Notation "'OKforall' x : a . b" :=
  (forallTK a (fun x:a => b)) (at level 30, right associativity).
Notation "'OKexists' x : a . b" :=
  (existsTK a (fun x:a => b)) (at level 30, right associativity).

Fixpoint to_OmegaK (x : Omega) {struct x} : OmegaK :=
  match x with
  | snat n => snatK n
  | sint i => sintK i
  | sbool b => sboolK b
  | arrow t1 t2 =>
    funcK
      (arrowK
        (tupleK 2
          (list_to_lambda OmegaK
            (cons (to_OmegaK t1) (cons (funcK (arrowK (to_OmegaK t2))) nil))
            (snatK 42))))
  | tuple t1 t2 => tupleK t1 (fun t:N => to_OmegaK(t2 t))
  | forallT k t =>
    funcK
      (forallTK k (fun t1:k => arrowK (funcK (arrowK (to_OmegaK (t t1))))))
  | existsT k t =>
    existsTK k (fun t1:k => to_OmegaK (t t1))
  end
.

Notation "'Kc' t " := (funcK (arrowK (to_OmegaK t))) (at level 20).

Definition cps_correctness (t: Omega) : OmegaK :=
  funcK (arrowK (funcK (arrowK (to_OmegaK t))))
.

Fixpoint to_closure (tk tenv: OmegaK) {struct tk} : OmegaK :=
  match tk with
  | snatK _
  | sintK _
  | sboolK _ => tk
  | arrowK t =>
    arrowK
      (tupleK 2
        (list_to_lambda OmegaK
          (cons tenv
            (cons (to_closure t
              (tupleK 0 (list_to_lambda OmegaK nil (snatK 42)))) nil))
          (snatK 42)))
  | tupleK t1 t2 => tupleK t1 (fun t'':N => to_closure (t2 t'') tenv)
  | forallTK k t => forallTK k (fun t1:k => to_closure (t t1) tenv)
  | existsTK k t => existsTK k (fun t1:k => to_closure (t t1) tenv)
  | funcK t =>
    existsTK OmegaK
      (fun t1:OmegaK =>
        tupleK 2
          (list_to_lambda OmegaK
            (cons (to_closure t t1) (cons t1 nil))
            (snatK 42)))
  end
.
