
open Pp

type identifier = Names.identifier

type ('term, 'expr, 'value) value_node =
        | CLVvar of int
        | CLVnat of 'term
        | CLVint of 'term
        | CLVbool of 'term
        | CLVpack of 'term * 'value * 'term
        | CLVtuple of 'value list
        | CLVfix of identifier * (identifier * 'term) list *
                    identifier * 'term * 'expr
        | CLVapp of 'value * 'term
and ('term, 'expr, 'value) expr_node =
        | CLEapp of 'value * 'value
        | CLEval of identifier * 'value * 'expr
        | CLEunpack of identifier * identifier * 'value * 'expr
        | CLEsel of identifier * 'term * 'value * 'value * 'expr
        | CLEunop of identifier * Clops.unop * 'value * 'expr
        | CLEbinop of identifier * Clops.binop * 'value * 'value * 'expr
        | CLEif of 'term * 'term *
                   'value * identifier * 'expr * identifier * 'expr
and value = (Prelude.term, expr, value) value_node
and expr = (Prelude.term, expr, value) expr_node

let clvvar i = CLVvar i
let clvnat t = CLVnat t
let clvint t = CLVint t
let clvbool t = CLVbool t
let clvpack t1 v t2 = CLVpack (t1, v, t2)
let clvtuple vl = CLVtuple vl
let clvfix id tl vid vt e = CLVfix (id, tl, vid, vt, e)
let clvapp v t = CLVapp (v, t)

let cleapp v1 v2 = CLEapp (v1, v2)
let cleval id v e = CLEval (id, v, e)
let cleunpack id1 id2 v e = CLEunpack (id1, id2, v, e)
let clesel id t v1 v2 e = CLEsel (id, t, v1, v2, e)
let cleunop id op v e = CLEunop (id, op, v, e)
let clebinop id op v1 v2 e = CLEbinop (id, op, v1, v2, e)
let cleif t1 t2 v id1 e1 id2 e2 = CLEif (t1, t2, v, id1, e1, id2, e2)

let cpsval_to_closval cpsv =
  let to_CL t1 t2 = Prelude.to_closure t1 t2 in
  let empty_tup = Prelude.tupleK [] in
  let lift_evars evars offset =
    let fmap index = index + offset in List.map fmap evars in
  let rec vwalk evars v =
    match v with
    | Cpsexpr.CVvar i -> clvvar (List.nth evars (i-1))
    | Cpsexpr.CVnat t -> clvnat t
    | Cpsexpr.CVint t -> clvint t
    | Cpsexpr.CVbool t -> clvbool t
    | Cpsexpr.CVpack (a, v, t) ->
      clvpack a (vwalk evars v) (to_CL t (empty_tup))
    | Cpsexpr.CVtuple vl -> clvtuple (List.map (vwalk evars) vl)
    | Cpsexpr.CVfix (x', defl, x, a, e, fevl, ftvl) ->
      let fevlen, ftvlen = List.length fevl, List.length ftvl in
      let defllen = List.length defl in
      let xname = Names.Name (Names.id_of_string "X") in
      let ftvlr = List.rev ftvl in
      (* helper function, create a list of integers : "ilist start max" *)
      let rec ilist i max = if i == max then [] else i :: ilist (i+1) max in
      (* create a list of free expr ids and a list of free expr types *)
      let fevids, fevt = List.split fevl in
      (* aenv *)
      let aenv_res = to_CL (Prelude.tupleK fevt) empty_tup in
      let aenv_body = (*Term.lift defllen*) aenv_res in
      (* create ax for pack *)
      let push_t (id, t) tt =
        Prelude.foralltK t (Term.mkLambda (Names.Name id, t, tt)) in
      let a'x =
        List.fold_right push_t defl
          (Prelude.arrowK
            (Prelude.tupleK
              [Term.mkRel (defllen + 1);
               to_CL ((*Term.liftn 1 (defllen + 1)*) a) empty_tup])) in
      let ax = Prelude.tupleK [a'x; Term.mkRel 1] in
      let ax_res =
        Prelude.existstK (Prelude.omegaK ())
           (Term.mkLambda (xname, Prelude.omegaK (), ax)) in
      (* create ax for body *)
      let push_t (id, t) tt =
        Prelude.foralltK t (Term.mkLambda (Names.Name id, t, tt)) in
      let a'x =
        List.fold_right push_t defl
          (Prelude.arrowK
            (Prelude.tupleK
              [Term.mkRel (defllen + 1);
               to_CL ((*Term.lift (defllen + 1)*) a) empty_tup])) in
      let ax = Prelude.tupleK [a'x; Term.mkRel 1] in
      let ax_body =
        Prelude.existstK (Prelude.omegaK ())
           (Term.mkLambda (xname, Prelude.omegaK (), ax)) in
      (* venv *)
      let venv =
        clvtuple
          (List.map (vwalk evars)
            (List.map Cpsexpr.cvvar (ilist 1 (fevlen + 1)))) in
      (* vcode *)
      let rec ftv_walk v ind =
        if ind == ftvlen then
          v
        else
          clvapp (ftv_walk v (ind+1)) (Term.mkRel (defllen + ind + 1)) in
      let x'v = ftv_walk (clvvar (fevlen + 3)) 0 in
      let evars = ilist 1 (fevlen + 3) in
      let vbottom =
        cleval x'
          (clvpack aenv_body (clvtuple [x'v; clvvar (fevlen + 1)]) ax_body)
          (clesel x
            (Prelude.lt_prop_t (Prelude.int_to_N 1) (Prelude.int_to_N 2))
            (clvvar (fevlen + 3)) (clvnat (Prelude.int_to_N 1))
              (ewalk evars e)) in
      let map_fevs ind rest =
        let ind_N = Prelude.int_to_N ind in
        clesel (List.nth fevids ind)
          (Prelude.lt_prop_t ind_N (Prelude.int_to_N fevlen))
          (clvvar (fevlen - ind)) (clvnat ind_N) rest in
      (* start here *)
      let vcode = 
        clvfix (Names.id_of_string "vfix") (ftvlr @ defl)
          (Names.id_of_string "xarg")
          (Prelude.tupleK [aenv_body; to_CL a empty_tup])
            (clesel
              (Names.id_of_string "xenv")
              (Prelude.lt_prop_t (Prelude.int_to_N 0) (Prelude.int_to_N 2))
              (clvvar 1) (clvnat (Prelude.int_to_N 0))
                (List.fold_right map_fevs (List.rev (ilist 0 fevlen))
                  vbottom))
       in
       clvpack
         aenv_res (clvtuple [vcode; venv]) ax_res
  and ewalk evars e =
    match e with
    | Cpsexpr.CEval (x, v, e) ->
      cleval x (vwalk evars v)
        (ewalk (1::(lift_evars evars 1)) e)
    | Cpsexpr.CEsel (x, a, v, v', e) ->
      clesel x a (vwalk evars v) (vwalk evars v')
        (ewalk (1::(lift_evars evars 1)) e)
    | Cpsexpr.CEunpack (x1, x2, v, e) ->
      cleunpack x1 x2 (vwalk evars v)
        (ewalk (1::(lift_evars evars 1)) e) 
    | Cpsexpr.CEbinop (x, op, v, v', e) ->
      clebinop x op (vwalk evars v) (vwalk evars v')
        (ewalk (1::(lift_evars evars 1)) e)
    | Cpsexpr.CEunop (x, op, v, e) ->
      cleunop x op (vwalk evars v)
        (ewalk (1::(lift_evars evars 1)) e)
    | Cpsexpr.CEif (b, a, v, x1, e1, x2, e2) ->
      cleif b a (vwalk evars v) x1 (ewalk evars e1) x2 (ewalk evars e2)
    | Cpsexpr.CEapp (v1, tl, v2) ->
      cleunpack
        (Names.id_of_string "xenv") (Names.id_of_string "xarg")
        (vwalk evars v1)
        (clesel (Names.id_of_string "xcode")
          (Prelude.lt_prop_t (Prelude.int_to_N 0) (Prelude.int_to_N 2))
          (clvvar 1) (clvnat (Prelude.int_to_N 0))
          (clesel (Names.id_of_string "xenv")
            (Prelude.lt_prop_t (Prelude.int_to_N 1) (Prelude.int_to_N 2))
            (clvvar 2) (clvnat (Prelude.int_to_N 1))
            (cleapp
              (List.fold_left clvapp (clvvar 2) tl) (*these need lifting*)
              (clvtuple [clvvar 1; vwalk (lift_evars evars 3) v2]))))
  in
    vwalk [] cpsv


let rec type_of_expr tctx ectx close =
  match close with
  | CLEapp (v1, v2) ->
    let v1_typ = type_of_value tctx ectx v1 in
    let v2_typ = type_of_value tctx ectx v2 in
    begin
    try
      let a = Prelude.dest_arrowK tctx v1_typ in
      if not (Prelude.eq_OmegaK tctx a v2_typ) then
        raise (Prelude.Type_Error (
          v 0 (str "Closure app arg type" ++ brk (0,0) ++
            hv 0 (
              hov 0 (Printer.pr_constr_env tctx v2_typ) ++ brk (1,0) ++
              str "!=" ++ brk (1, 0) ++
              hov 0 (Printer.pr_constr_env tctx a)))))
    with Prelude.Bad_type _ -> raise (Failure "Closure app func type")
    end
  | CLEval (x, v, e) ->
    let v_typ = type_of_value tctx ectx v in
    let ectx' = Environ.push_rel (Names.Name x, None, v_typ) ectx in
    type_of_expr tctx ectx' e
  | CLEsel (x, a, v, v', e) ->
    begin
    try
      let v_typ = type_of_value tctx ectx v in
      let (a'', b) = Prelude.dest_tupleK tctx v_typ in
      let v'_typ = type_of_value tctx ectx v' in
      let (a', is_int) = Prelude.dest_arithmeticK tctx v'_typ in
      if not is_int then
        let cond = Prelude.lt_prop_t a' a'' in
        if Prelude.term_equal_env tctx cond a then
          let ba' = Term.mkApp (b, [| a' |]) in
          let ectx' = Environ.push_rel (Names.Name x, None, ba') ectx in
          type_of_expr tctx ectx' e
        else
          raise (Prelude.Type_Error (
            Pp.v 0 (str "Closure sel expr type of A" ++ brk (0,0) ++
              hv 0 (
                hov 0 (Printer.pr_constr_env tctx cond) ++ brk (1,0) ++
                str "!=" ++ brk (1, 0) ++
                hov 0 (Printer.pr_constr_env tctx a)))))
      else
        raise (Failure "Closure sel expr with Z index")
    with Prelude.Bad_type _ -> raise (Failure "Closure sel expr type")
    end
  | CLEunpack (x1, x2, v, e) ->
    begin
    try
      let v_typ = type_of_value tctx ectx v in
      let (b, f) = Prelude.dest_existsK tctx v_typ in
      let ectx', tctx' =
        Prelude.tctx_push (Names.Name x1, None, b) ectx tctx in
      let a = Term.mkApp (f, [| Term.mkRel 1 |]) in
      let ectx' = Environ.push_rel (Names.Name x2, None, a) ectx' in
      type_of_expr tctx' ectx' e
    with Prelude.Bad_type _ -> raise (Failure "Closure open expr type")
    end
  | CLEunop (x, op, v, e) ->
    begin
    try
      let v_typ = type_of_value tctx ectx v in
      let x_typ =
        match op with
        | Clops.UOnot ->
          Prelude.sboolK
            (Prelude.unop_bool op (Prelude.dest_boolK tctx v_typ)) in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      type_of_expr tctx ectx' e
    with Prelude.Bad_type _ -> raise (Failure "CPS unop expr type")
    end
  | CLEbinop (x, op, v, v', e) ->
    begin
    try
      let v_typ = type_of_value tctx ectx v in
      let v'_typ = type_of_value tctx ectx v' in
      let x_typ =
        match op with
        | Clops.Blogicop _ ->
          Prelude.sboolK
            (Prelude.binop_bool op
              (Prelude.dest_boolK tctx v_typ) (Prelude.dest_boolK tctx v'_typ))
        | Clops.Barithop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sintK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.snatK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end
        | Clops.Brelop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sboolK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.sboolK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      type_of_expr tctx ectx' e
    with Prelude.Bad_type _ -> raise (Failure "Closure binop expr type")
    end
  | CLEif (b, a, v, x1, e1, x2, e2) ->
    let btos = Term.mkArrow (Prelude.cterm "bool") Term.mkProp in
    let b_typ = Prelude.term_type tctx b in
    if Prelude.term_equal_env tctx b_typ btos then
      try
        let v_typ = type_of_value tctx ectx v in
        let a' = Prelude.dest_boolK tctx v_typ in
        let ba' = Term.mkApp (b, [| a' |]) in
        let a_typ = Prelude.term_type tctx a in
        if Prelude.term_equal_env tctx a_typ ba' then
          let b_tru =
            Term.mkApp (b, [| Prelude.cterm "true" |]) in
          let ectxbt, tctxbt =
            Prelude.tctx_push (Names.Name x1, None, b_tru) ectx tctx in
          type_of_expr tctxbt ectxbt e1;
          let b_fls =
            Term.mkApp (b, [| Prelude.cterm "false" |]) in
          let ectxbf, tctxbf =
            Prelude.tctx_push (Names.Name x2, None, b_fls) ectx tctx in
          type_of_expr tctxbf ectxbf e2
        else
          raise (Failure "Closure if expr type A")
      with Prelude.Bad_type _ ->
        raise (Failure "Closude if expr type v")
    else
      raise (Failure "Closure if expr type B")
    
and type_of_value tctx ectx closv =
  match closv with
  | CLVvar i ->
    begin
    try
      let (_, _, t) = Environ.lookup_rel i ectx in t
    with _ -> raise (Prelude.Internal_Error "closure typing index")
    end
  | CLVnat t -> Prelude.snatK t
  | CLVint t -> Prelude.sintK t
  | CLVbool t -> Prelude.sboolK t
  | CLVpack (a, v, t) ->
    let (b, t') = Prelude.dest_existsK tctx t in
    let a_typ = Prelude.term_type tctx a in
    if Prelude.term_equal_env tctx a_typ b then
      let v_typ = type_of_value tctx ectx v in
      let t'a = Term.mkApp (t', [| a |]) in
        if Prelude.eq_OmegaK tctx v_typ t'a then
          t
        else
          raise (Prelude.Type_Error (
            Pp.v 0 (str "Closure pack expr type" ++ brk (0,0) ++
              hv 0 (
                hov 0 (Printer.pr_constr_env tctx v_typ) ++ brk (1,0) ++
                str "!=" ++ brk (1, 0) ++
                hov 0 (Printer.pr_constr_env tctx t'a)))))
    else
      raise (Failure "Closure pack A type")
  | CLVtuple vl ->
    Prelude.tupleK (List.map (type_of_value tctx ectx) vl)
  | CLVapp (v, a) ->
    let v_typ = type_of_value tctx ectx v in
    begin
    try
      let (a', t) = Prelude.dest_forallK tctx v_typ in
      let a_typ = Prelude.term_type tctx a in
      if Prelude.term_equal_env tctx a' a_typ then
        Term.mkApp (t, [| a |])
      else
        raise (Failure "Closure tapp forall arg")
    with Prelude.Bad_type _ -> raise (Failure "Closure tapp func not forall")
    end
  | CLVfix (x', defl, x, a, e) ->
    let ectx, tctx = Environ.empty_env, Global.env () in
    let tl = List.map snd defl in
    let tl_typ = List.map (Prelude.term_type tctx) tl in
    if List.for_all Term.isSort tl_typ then
      (* Create a' *)
      let push_ex (id, t) tt =
        Prelude.foralltK t (Term.mkLambda (Names.Name id, t, tt)) in
      let a' = List.fold_right push_ex defl (Prelude.arrowK a) in
      (* create the extended environments *)
      let push_tvar tctx (id, t) =
        Environ.push_rel (Names.Name id, None, t) tctx in
      let tctx' = List.fold_left push_tvar tctx defl in
      let a_typ = Prelude.term_type tctx' a in
      if Prelude.term_equal_env tctx' a_typ (Prelude.omegaK ()) then
        let defl_len = List.length defl in
        let (ectx', a'') =
          if defl_len > 0 then
            let lift_decl (id, body, t) = (id, body, Term.lift defl_len t) in
            let e_rel_ctx = List.map lift_decl (Environ.rel_context ectx) in
            (Environ.push_rel_context e_rel_ctx Environ.empty_env,
             Term.lift defl_len a')
          else
            (ectx, a') in
        let ectx' = Environ.push_rel (Names.Name x', None, a'') ectx' in
        let ectx' = Environ.push_rel (Names.Name x, None, a) ectx' in
        let _ = type_of_expr tctx' ectx' e in
          a'
      else
        raise (Failure "Closure fix A type not OmegaK")
    else
      raise (Failure "Closure fix types not Sort")

let pr_value tctx ectx cpsv =
  let pr_tt = Printer.pr_constr in
  let pr_tt_env = Printer.pr_constr_env in
  let pr_id id = str (Names.string_of_id id) in
  let rec ewalk tctx ectx e =
    match e with
    | CLEapp (v1, v2) ->
      hov 0 (
        hov 1 (str "(" ++ vwalk tctx ectx v1 ++ str ")") ++ brk (1,0) ++
        hov 1 (str "(" ++ vwalk tctx ectx v2 ++ str ")"))
    | CLEval (x, v, e) ->
      let ectx' = ectx in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk (1,2) ++
        vwalk tctx ectx v ++ brk(1,0) ++ str "in" ++ brk (1,2) ++
        ewalk tctx ectx' e)
    | CLEunpack (x1, x2, v, e) ->
      let ectx', tctx' = ectx, tctx in
      hov 0 (
        str "let " ++ pr_id x1 ++ str ", " ++ pr_id x2 ++ str " = open" ++
        brk(1,2) ++ vwalk tctx ectx v ++ brk(1,0) ++ str "in" ++ brk(1,2) ++
        ewalk tctx' ectx' e)
    | CLEsel (x, a, v, v', e) ->
      let ectx' = ectx in
      hov 0 (
        str "let " ++ pr_id x ++ str " = sel [" ++ brk(1,2) ++
        pr_tt_env tctx a ++ str "]" ++ brk(1,2) ++
        str "(" ++ vwalk tctx ectx v ++ str "," ++ brk(1,2) ++
        vwalk tctx ectx v' ++ str ")" ++ brk(1,0) ++
        str "in" ++ brk(1,2) ++ ewalk tctx ectx' e)
    | CLEunop (x, op, v, e) ->
      let ectx' = ectx in
      let op_str = Clops.pretty_unop op in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk(1,2) ++
        str op_str ++ hov (String.length op_str) (vwalk tctx ectx v) ++
        brk(1,0) ++ str "in" ++ brk(1,2) ++ ewalk tctx ectx' e)
    | CLEbinop (x, op, v, v', e) ->
      let ectx' = ectx in
      let op_str = Clops.pretty_binop op in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk(1,2) ++
        vwalk tctx ectx v ++ str op_str ++ brk(1,2) ++
        hov 0 (vwalk tctx ectx v') ++ brk(1,0) ++ str "in" ++ brk(1,2) ++
        ewalk tctx ectx' e)
    | CLEif (b, a, v, x1, e1, x2, e2) -> 
      let ectxbt, tctxbt = ectx, tctx in
      let ectxbf, tctxbf = ectx, tctx in
      hov 2 (
        str "if [" ++ pr_tt_env tctx b ++ str ", " ++ pr_tt_env tctx a ++
        str "]" ++ brk (1,0) ++
        str "(" ++ vwalk tctx ectx v ++ str "," ++ brk (1,0) ++
        pr_id x1 ++ str "." ++ ewalk tctxbt ectxbt e1 ++ str "," ++
        brk (1,0) ++ pr_id x2 ++ str "." ++ ewalk tctxbf ectxbf e2 ++ str ")")
  and vwalk tctx ectx v =
    match v with
    | CLVvar i ->
      hov 0 (str "EVAR_" ++ str (string_of_int i))
    | CLVnat t ->
      hov 0 (str "cnatK [" ++ pr_tt t ++ str "]")
    | CLVint t ->
      hov 0 (str "cintK [" ++ pr_tt t ++ str "]")
    | CLVbool t ->
      hov 0 (str "cboolK [" ++ pr_tt t ++ str "]")
    | CLVpack (a, v, t) ->
      hov 0 (
        str "pack" ++ brk (1,2) ++
        str "(" ++ pr_tt_env tctx a ++ str "," ++ brk (1,2) ++
        vwalk tctx ectx v ++ str ")" ++ brk (1,0) ++ str "as" ++ brk (1,2) ++
        pr_tt_env tctx t)
    | CLVtuple vl ->
      let rec do_elem l =
        match l with
        | [] -> str ""
        | el :: [] -> vwalk tctx ectx el
        | el :: rest ->
          vwalk tctx ectx el ++ str "," ++ brk (1,0) ++ do_elem rest in
      hov 1 (str "{" ++ do_elem vl ++ str "}")
    | CLVfix (x', defl, x, a, e) ->
      let ectx' = ectx in
      let tctx' = tctx in
      let rec do_elem l =
        match l with
        | [] -> str ""
        | (id, t) :: [] ->
          pr_id id ++ str ":" ++ pr_tt_env tctx t
        | (id, t) :: rest ->
          pr_id id ++ str ":" ++ pr_tt_env tctx t ++ str "," ++ brk(1,0) ++
          do_elem rest in
      hov 2 (
        str "fix " ++ pr_id x' ++ str " [" ++ brk(1,0) ++
        hov 0 (do_elem defl) ++ str "]" ++ brk(1,0) ++
        str "(" ++ pr_id x ++ str ":" ++ pr_tt_env tctx' a ++ str ")." ++
        brk(1,0) ++ ewalk tctx' ectx' e)
    | CLVapp (v, t) ->
      hov 0 (
        hov 1 (str "(" ++ vwalk tctx ectx v ++ str ")") ++ brk (1,0) ++
        hov 1 (str "[" ++ pr_tt_env tctx t ++ str "]"))
  in
  vwalk tctx ectx cpsv

