open Format
open Pp

type identifier = Names.identifier

type ('term, 'expr, 'value) value_node =
        | CVvar of int
        | CVnat of 'term
        | CVint of 'term
        | CVbool of 'term
        | CVpack of 'term * 'value * 'term
        | CVtuple of 'value list
        | CVfix of identifier * (identifier * 'term) list *
                   identifier * 'term * 'expr *
                   (identifier * 'term) list * (identifier * 'term) list
and ('term, 'expr, 'value) expr_node =
        | CEapp of 'value * 'term list * 'value
        | CEval of identifier * 'value * 'expr
        | CEunpack of identifier * identifier * 'value * 'expr
        | CEsel of identifier * 'term * 'value * 'value * 'expr
        | CEunop of identifier * Clops.unop * 'value * 'expr
        | CEbinop of identifier * Clops.binop * 'value * 'value * 'expr
        | CEif of 'term * 'term *
                  'value * identifier * 'expr * identifier * 'expr
and value = (Prelude.term, expr, value) value_node
and expr = (Prelude.term, expr, value) expr_node

let cvvar i = CVvar i
let cvnat t = CVnat t
let cvint t = CVint t
let cvbool t = CVbool t
let cvpack t1 v t2 = CVpack (t1, v, t2)
let cvtuple vl = CVtuple vl
let cvfix id tl vid vt e fev ftv = CVfix (id, tl, vid, vt, e, fev, ftv)

let ceapp v1 tl v2 = CEapp (v1, tl, v2)
let ceval id v e = CEval (id, v, e)
let ceunpack id1 id2 v e = CEunpack (id1, id2, v, e)
let cesel id t v1 v2 e = CEsel (id, t, v1, v2, e)
let ceunop id op v e = CEunop (id, op, v, e)
let cebinop id op v1 v2 e = CEbinop (id, op, v1, v2, e)
let ceif t1 t2 v id1 e1 id2 e2 = CEif (t1, t2, v, id1, e1, id2, e2)

let fresh =
  let index = ref 0 in
  function x ->
    incr index;
    Names.id_of_string ("_" ^ x ^ "_" ^ (string_of_int !index))

let aexpr_to_cps ae =
  let to_K t = Prelude.to_omegaK t in
  let to_FK t = Prelude.funcK (Prelude.arrowK (to_K t)) in
  let lambdaK id t e =
    cvfix (fresh "lk") [] id t e [] [] in
  let lift_evars evars offset =
    let fmap index = index + offset in List.map fmap evars in
  let rec ae_walk evars ae =
    let (e, et) = ae in
    match e with
    | Expr.Evar i ->
      let new_index = List.nth (lift_evars evars 2) (i-1) in
      lambdaK (fresh "kvar") (to_FK et) (ceapp (cvvar 1) [] (cvvar new_index))
    | Expr.Ecnat i ->
      lambdaK (fresh "knat") (to_FK et) (ceapp (cvvar 1) [] (cvnat i))
    | Expr.Ecint i ->
      lambdaK (fresh "kint") (to_FK et) (ceapp (cvvar 1) [] (cvint i))
    | Expr.Ecbool b ->
      lambdaK (fresh "kbool") (to_FK et) (ceapp (cvvar 1) [] (cvbool b))
    | Expr.Elambda _
    | Expr.Epoly _ ->
      lambdaK (fresh "kf") (to_FK et)
        (ceapp (cvvar 1) [] (fval_walk (lift_evars evars 2) ae))
    | Expr.Efix (x', _, fv) ->
      begin
        match fv with
        | (Expr.Elambda (x, a, e), _) ->
          let zero = Prelude.int_to_N 0 in
          let one = Prelude.int_to_N 1 in
          let two = Prelude.int_to_N 2 in
          lambdaK (fresh "kfixlam") (to_FK et)
            (ceapp
              (cvvar 1) []
              (cvfix x' [] (fresh "xarg")
                (Prelude.tupleK [to_K a; to_FK (snd e)])
                (cesel x (Prelude.lt_prop_t zero two)
                  (cvvar 1) (cvnat zero)
                  (cesel (fresh "c") (Prelude.lt_prop_t one two)
                    (cvvar 2) (cvnat one)
                    (ceapp
                      (ae_walk (2::4::(lift_evars evars 6)) e) []
                      (cvvar 1)))) [] []))
        | (Expr.Epoly (x, a, f), _) ->
          lambdaK (fresh "kfixpol") (to_FK et)
            (ceapp
              (cvvar 1) []
              (cvfix x' [(x, a)] (fresh "c")
                (to_FK (snd f))
                (ceapp (cvvar 1) []
                  (fval_walk (2::(lift_evars evars 4)) f)) [] []))
        | _ -> raise (Failure "Fix with non-function expr!")
      end
    | Expr.Eapp (e1, e2) ->
      lambdaK (fresh "kapp") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e1) []
          (lambdaK (fresh "x1") (to_K (snd e1))
            (ceapp
              (ae_walk (lift_evars evars 4) e2) []
              (lambdaK (fresh "x2") (to_K (snd e2))
                (ceapp
                  (cvvar 3) []
                  (cvtuple [cvvar 1; cvvar 5]))))))
    | Expr.Etapp (e, a) ->
      lambdaK (fresh "ktapp") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e) []
          (lambdaK (fresh "x") (to_K (snd e)) (ceapp (cvvar 1) [a] (cvvar 3))))
    | Expr.Etuple ael ->
      let rec map aelist index res_tuple =
        match aelist with
        | [] -> ceapp (cvvar index) [] (cvtuple res_tuple)
        | hd :: tl ->
          ceapp
            (ae_walk (lift_evars evars (index+1)) hd) []
            (lambdaK (fresh ("x"^(string_of_int ((index - 1)/2))))
              (to_K (snd hd))
              (map tl (index + 2) (cvvar index::res_tuple))) in
      lambdaK (fresh "ktup") (to_FK et) (map ael 1 [])
    | Expr.Esel (a, e1, e2) ->
      lambdaK (fresh "ksel") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e1) []
          (lambdaK (fresh "x1") (to_K (snd e1))
            (ceapp
              (ae_walk (lift_evars evars 4) e2) []
              (lambdaK (fresh "x2") (to_K (snd e2))
                (cesel (fresh "c") a (cvvar 3) (cvvar 1)
                  (ceapp (cvvar 6) [] (cvvar 1)))))))
    | Expr.Epack (a, e, t) ->
      lambdaK (fresh "kpack") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e) []
          (lambdaK (fresh "x") (to_K (snd e))
            (ceapp (cvvar 3) [] (cvpack a (cvvar 1) (to_K t)))))
    | Expr.Eunpack (e1, x1, x2, e2) ->
      lambdaK (fresh "kopen") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e1) []
          (lambdaK (fresh "x1") (to_K (snd e1))
            (ceunpack x1 x2 (cvvar 1)
              (ceapp (ae_walk (1::(lift_evars evars 5)) e2) [] (cvvar 4)))))
    | Expr.Ebinop (op, e1, e2) ->
      lambdaK (fresh "kbop") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e1) []
          (lambdaK (fresh "x1") (to_K (snd e1))
            (ceapp
              (ae_walk (lift_evars evars 4) e2) []
              (lambdaK (fresh "x2") (to_K (snd e2))
                (cebinop (fresh "c") op (cvvar 3) (cvvar 1)
                  (ceapp (cvvar 6) [] (cvvar 1)) )))))
    | Expr.Eunop (op, e) ->
      lambdaK (fresh "kunop") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e) []
          (lambdaK (fresh "x") (to_K (snd e))
            (ceunop (fresh "c") op (cvvar 1) (ceapp (cvvar 4) [] (cvvar 1)))))
    | Expr.Eif (b, a, e, x1, e1, x2, e2) ->
      lambdaK (fresh "kif") (to_FK et)
        (ceapp
          (ae_walk (lift_evars evars 2) e) []
          (lambdaK (fresh "x") (to_K (snd e))
            (ceif b a (cvvar 1)
              x1 (ceapp (ae_walk (lift_evars evars 4) e1) [] (cvvar 3))
              x2 (ceapp (ae_walk (lift_evars evars 4) e2) [] (cvvar 3)))))
  and fval_walk evars ae =
    match ae with
    | (Expr.Elambda (x, a, e), _) ->
      let zero = Prelude.int_to_N 0 in
      let one = Prelude.int_to_N 1 in
      let two = Prelude.int_to_N 2 in
      lambdaK (fresh "xarg") (Prelude.tupleK [to_K a; to_FK (snd e)])
        (cesel x (Prelude.lt_prop_t zero two)
          (cvvar 1) (cvnat zero)
          (cesel (fresh "c") (Prelude.lt_prop_t one two)
            (cvvar 2) (cvnat one)
            (ceapp (ae_walk (2::(lift_evars evars 4)) e) [] (cvvar 1))))
    | (Expr.Epoly (x, a, f), _) ->
      cvfix (fresh "fpoly") [(x, a)] (fresh "kpoly")
        (to_FK (snd f))
        (ceapp
          (cvvar 1) []
          (fval_walk (lift_evars evars 2) f)) [] []
    | _ -> raise (Failure "fval with non-function expr!")
  in
    ae_walk [] ae


let rec type_of_expr tctx ectx cpse =
  match cpse with
  | CEapp (v', tl, v) ->
    begin
    try
      let (nv', v'_typ) = type_of_value tctx ectx v' in
      let v'_typ = Prelude.dest_funcK tctx v'_typ in
      let dest_f v ael =
        let b1 = Prelude.term_type tctx ael in
        let (b2, f) = Prelude.dest_forallK tctx v in
        if Prelude.term_equal_env tctx b1 b2 then
          Term.mkApp (f, [| ael |])
        else
          raise (Failure "CPS app expr type arguments") in
      let a = List.fold_left dest_f v'_typ tl in
      let a = Prelude.dest_arrowK tctx a in
      let (nv, v_typ) = type_of_value tctx ectx v in
      if not (Prelude.eq_OmegaK tctx a v_typ) then
        raise (Prelude.Type_Error (
          Pp.v 0 (str "app arg type" ++ brk (0,0) ++
            hv 0 (
              str "Expression :" ++ brk (1, 2) ++
              hov 0 (pr_value tctx ectx v' ++ brk (1,0) ++
                pr_value tctx ectx v) ++ brk (1,0) ++
              str "Type:" ++ brk (1, 0) ++
              hov 0 (Printer.pr_constr_env tctx v_typ) ++ brk (1,0) ++
              str "should be" ++ brk (1, 0) ++
              hov 0 (Printer.pr_constr_env tctx a)))))
      else
        ceapp nv' tl nv
    with Prelude.Bad_type _ -> raise (Failure "CPS app first expression")
    end
  | CEval (x, v, e) ->
    let (nv, v_typ) = type_of_value tctx ectx v in
    let ectx' = Environ.push_rel (Names.Name x, None, v_typ) ectx in
    let ne = type_of_expr tctx ectx' e in
      ceval x nv ne
  | CEsel (x, a, v, v', e) ->
    begin
    try
      let (nv, v_typ) = type_of_value tctx ectx v in
      let (a'', b) = Prelude.dest_tupleK tctx v_typ in
      let (nv', v'_typ) = type_of_value tctx ectx v' in
      let (a', is_int) = Prelude.dest_arithmeticK tctx v'_typ in
      if not is_int then
        let cond = Prelude.lt_prop_t a' a'' in
        if Prelude.term_equal_env tctx cond a then
          let ba' = Term.mkApp (b, [| a' |]) in
          let ectx' = Environ.push_rel (Names.Name x, None, ba') ectx in
          let ne = type_of_expr tctx ectx' e in
            cesel x a nv nv' ne
        else
          raise (Failure "CPS sel expr type of A")
      else
        raise (Failure "CPS sel expr with Z index")
    with Prelude.Bad_type _ -> raise (Failure "CPS sel expr type")
    end
  | CEunpack (x1, x2, v, e) ->
    begin
    try
      let (nv, v_typ) = type_of_value tctx ectx v in
      let (b, f) = Prelude.dest_existsK tctx v_typ in
      let ectx', tctx' =
        Prelude.tctx_push (Names.Name x1, None, b) ectx tctx in
      let a = Term.mkApp (f, [| Term.mkRel 1 |]) in
      let ectx' = Environ.push_rel (Names.Name x2, None, a) ectx' in
      let ne = type_of_expr tctx' ectx' e in
        ceunpack x1 x2 nv ne
    with Prelude.Bad_type _ -> raise (Failure "CPS open expr type")
    end
  | CEunop (x, op, v, e) ->
    begin
    try
      let (nv, v_typ) = type_of_value tctx ectx v in
      let x_typ =
        match op with
        | Clops.UOnot ->
          Prelude.sboolK
            (Prelude.unop_bool op (Prelude.dest_boolK tctx v_typ)) in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      let ne = type_of_expr tctx ectx' e in
        ceunop x op nv ne
    with Prelude.Bad_type _ -> raise (Failure "CPS unop expr type")
    end
  | CEbinop (x, op, v, v', e) ->
    begin
    try
      let (nv, v_typ) = type_of_value tctx ectx v in
      let (nv', v'_typ) = type_of_value tctx ectx v' in
      let x_typ =
        match op with
        | Clops.Blogicop _ ->
          Prelude.sboolK
            (Prelude.binop_bool op
              (Prelude.dest_boolK tctx v_typ) (Prelude.dest_boolK tctx v'_typ))
        | Clops.Barithop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sintK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.snatK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end
        | Clops.Brelop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sboolK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.sboolK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      let ne = type_of_expr tctx ectx' e in
        cebinop x op nv nv' ne
    with Prelude.Bad_type _ -> raise (Failure "CPS binop expr type")
    end
  | CEif (b, a, v, x1, e1, x2, e2) ->
    let btos = Term.mkArrow (Prelude.cterm "bool") Term.mkProp in
    let b_typ = Prelude.term_type tctx b in
    if Prelude.term_equal_env tctx b_typ btos then
      try
        let (nv, v_typ) = type_of_value tctx ectx v in
        let a' = Prelude.dest_boolK tctx v_typ in
        let ba' = Term.mkApp (b, [| a' |]) in
        let a_typ = Prelude.term_type tctx a in
        if Prelude.term_equal_env tctx a_typ ba' then
          let b_tru =
            Term.mkApp (b, [| Prelude.cterm "true" |]) in
          let ectxbt, tctxbt =
            Prelude.tctx_push (Names.Name x1, None, b_tru) ectx tctx in
          let ne1 = type_of_expr tctxbt ectxbt e1 in
          let b_fls =
            Term.mkApp (b, [| Prelude.cterm "false" |]) in
          let ectxbf, tctxbf =
            Prelude.tctx_push (Names.Name x2, None, b_fls) ectx tctx in
          let ne2 = type_of_expr tctxbf ectxbf e2 in
            ceif b a nv x1 ne1 x2 ne2
        else
          raise (Failure "CPS if expr type A")
      with Prelude.Bad_type _ ->
        raise (Failure "CPS if expr type v")
    else
      raise (Failure "CPS if expr type B")
    
and type_of_value tctx ectx cpsv =
  match cpsv with
  | CVvar i -> let (_, _, t) = Environ.lookup_rel i ectx in (cvvar i, t)
  | CVnat t -> (cvnat t, Prelude.snatK t)
  | CVint t -> (cvint t, Prelude.sintK t)
  | CVbool t -> (cvbool t, Prelude.sboolK t)
  | CVpack (a, v, t) ->
    let (b, t') = Prelude.dest_existsK tctx t in
    let a_typ = Prelude.term_type tctx a in
    if Prelude.term_equal_env tctx a_typ b then
      let (nv, v_typ) = type_of_value tctx ectx v in
      let t'a = Term.mkApp (t', [| a |]) in
        if Prelude.eq_OmegaK tctx v_typ t'a then
          (cvpack a nv t, t)
        else
          raise (Failure "CPS pack expr type")
    else
      raise (Failure "CPS pack A type")
  | CVtuple vl ->
    let (nvl, vlt) = List.split (List.map (type_of_value tctx ectx) vl) in
      (cvtuple nvl, Prelude.tupleK vlt)
  | CVfix (x', defl, x, a, e, _, _) ->
    let env_to_list e =
      List.map
        (fun (i,_,t) -> (Prelude.id_of_name i, t)) (Environ.rel_context e) in
    let tl = List.map snd defl in
    let tl_typ = List.map (Prelude.term_type tctx) tl in
    if List.for_all Term.isSort tl_typ then (* whnf is needed here probably *)
      (* Create a' *)
      let push_ex (id, t) tt =
        Prelude.foralltK t (Term.mkLambda (Names.Name id, t, tt)) in
      let a' = List.fold_right push_ex defl (Prelude.arrowK a) in
      let a' = Prelude.funcK a' in
      (* create the extended environments *)
      let push_tvar tctx (id, t) =
        Environ.push_rel (Names.Name id, None, t) tctx in
      let tctx' = List.fold_left push_tvar tctx defl in
      let a_typ = Prelude.term_type tctx' a in
      if Prelude.term_equal_env tctx' a_typ (Prelude.omegaK ()) then
        let defl_len = List.length defl in
        let (ectx', a'') =
          if defl_len > 0 then
            let lift_decl (id, body, t) = (id, body, Term.lift defl_len t) in
            let e_rel_ctx = List.map lift_decl (Environ.rel_context ectx) in
            (Environ.push_rel_context e_rel_ctx Environ.empty_env,
             Term.lift defl_len a')
          else
            (ectx, a') in
        let ectx' = Environ.push_rel (Names.Name x', None, a'') ectx' in
        let ectx' = Environ.push_rel (Names.Name x, None, a) ectx' in
        let ne = type_of_expr tctx' ectx' e in
          (cvfix x' defl x a ne (env_to_list ectx) (env_to_list tctx), a')
      else
        raise (Failure "CPS fix A type")
    else
      raise (Failure "CPS fix types not Sort")

and pr_value tctx ectx cpsv =
  let pr_tt = Printer.pr_constr in
  let pr_tt_env = Printer.pr_constr_env in
  let pr_id id = str (Names.string_of_id id) in
  let rec ewalk tctx ectx e =
    match e with
    | CEapp (v', tl, v) ->
      let rec do_elem l =
        match l with
        | [] -> str ""
        | el :: [] -> pr_tt_env tctx el
        | el :: rest ->
          pr_tt_env tctx el ++ str "," ++ brk(1,0) ++ do_elem rest in
      hov 0 (
        hov 1 (str "(" ++ vwalk tctx ectx v' ++ str ")") ++ brk (1,0) ++
        hov 1 (str "[" ++ do_elem tl ++ str "]") ++ brk (1,0) ++
        hov 1 (str "(" ++ vwalk tctx ectx v ++ str ")"))
    | CEval (x, v, e) ->
      let v_typ = snd (type_of_value tctx ectx v) in
      let ectx' = Environ.push_rel (Names.Name x, None, v_typ) ectx in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk (1,2) ++
        vwalk tctx ectx v ++ brk(1,0) ++ str "in" ++ brk (1,2) ++
        ewalk tctx ectx' e)
    | CEunpack (x1, x2, v, e) ->
      let v_typ = snd (type_of_value tctx ectx v) in
      let (b, t) = Prelude.dest_existsK tctx v_typ in
      let ectx', tctx' = Prelude.tctx_push (Names.Name x1, None, b) ectx tctx in
      let a = Term.mkApp (t, [| Term.mkRel 1 |]) in
      let ectx' = Environ.push_rel (Names.Name x2, None, a) ectx' in
      hov 0 (
        str "let " ++ pr_id x1 ++ str ", " ++ pr_id x2 ++ str " = open" ++
        brk(1,2) ++ vwalk tctx ectx v ++ brk(1,0) ++ str "in" ++ brk(1,2) ++
        ewalk tctx' ectx' e)
    | CEsel (x, a, v, v', e) ->
      let v_typ = snd (type_of_value tctx ectx v) in
      let b = snd (Prelude.dest_tupleK tctx v_typ) in
      let v'_typ = snd (type_of_value tctx ectx v') in
      let a' = fst (Prelude.dest_arithmeticK tctx v'_typ) in
      let ba' = Term.mkApp (b, [| a' |]) in
      let ectx' = Environ.push_rel (Names.Name x, None, ba') ectx in
      hov 0 (
        str "let " ++ pr_id x ++ str " = sel [" ++ brk(1,2) ++
        pr_tt_env tctx a ++ str "]" ++ brk(1,2) ++
        str "(" ++ hov 1 (vwalk tctx ectx v) ++ str "," ++ brk(1,2) ++
        vwalk tctx ectx v' ++ str ")" ++ brk(1,0) ++
        str "in" ++ brk(1,2) ++ ewalk tctx ectx' e)
    | CEunop (x, op, v, e) ->
      let v_typ = snd (type_of_value tctx ectx v) in
      let x_typ =
        match op with
        | Clops.UOnot ->
          Prelude.sboolK
            (Prelude.unop_bool op (Prelude.dest_boolK tctx v_typ)) in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      let op_str = Clops.pretty_unop op in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk(1,2) ++
        str op_str ++ hov (String.length op_str) (vwalk tctx ectx v) ++
        brk(1,0) ++ str "in" ++ brk(1,2) ++ ewalk tctx ectx' e)
    | CEbinop (x, op, v, v', e) ->
      let v_typ = snd (type_of_value tctx ectx v) in
      let v'_typ = snd (type_of_value tctx ectx v') in
      let x_typ =
        match op with
        | Clops.Blogicop _ ->
          Prelude.sboolK
            (Prelude.binop_bool op
              (Prelude.dest_boolK tctx v_typ) (Prelude.dest_boolK tctx v'_typ))
        | Clops.Barithop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sintK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.snatK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end
        | Clops.Brelop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sboolK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.sboolK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      let op_str = Clops.pretty_binop op in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk(1,2) ++
        vwalk tctx ectx v ++ brk(1,2) ++ str op_str ++ brk(1,2) ++
        hov 0 (vwalk tctx ectx v') ++ brk(1,0) ++ str "in" ++ brk(1,2) ++
        ewalk tctx ectx' e)
    | CEif (b, a, v, x1, e1, x2, e2) -> 
      let b_tru = Term.mkApp (b, [| Prelude.cterm "true" |]) in
      let ectxbt, tctxbt =
        Prelude.tctx_push (Names.Name x1, None, b_tru) ectx tctx in
      let b_fls = Term.mkApp (b, [| Prelude.cterm "false" |]) in
      let ectxbf, tctxbf =
        Prelude.tctx_push (Names.Name x2, None, b_fls) ectx tctx in
      hov 2 (
        str "if [" ++ pr_tt_env tctx b ++ str ", " ++ pr_tt_env tctx a ++
        str "]" ++ brk (1,0) ++
        str "(" ++ vwalk tctx ectx v ++ str "," ++ brk (1,0) ++
        pr_id x1 ++ str "." ++ ewalk tctxbt ectxbt e1 ++ str "," ++
        brk (1,0) ++ pr_id x2 ++ str "." ++ ewalk tctxbf ectxbf e2 ++ str ")")
  and vwalk tctx ectx v =
    match v with
    | CVvar i ->
      let (var, _, _) = Environ.lookup_rel i ectx in
      hov 0 (str (Prelude.string_of_name var))
    | CVnat t ->
      hov 0 (str "cnatK [" ++ pr_tt t ++ str "]")
    | CVint t ->
      hov 0 (str "cintK [" ++ pr_tt t ++ str "]")
    | CVbool t ->
      hov 0 (str "cboolK [" ++ pr_tt t ++ str "]")
    | CVpack (a, v, t) ->
      hov 0 (
        str "pack" ++ brk (1,2) ++
        str "(" ++ pr_tt_env tctx a ++ str "," ++ brk (1,2) ++
        vwalk tctx ectx v ++ str ")" ++ brk (1,0) ++ str "as" ++ brk (1,2) ++
        pr_tt_env tctx t)
    | CVtuple vl ->
      let rec do_elem l =
        match l with
        | [] -> str ""
        | el :: [] -> vwalk tctx ectx el
        | el :: rest ->
          vwalk tctx ectx el ++ str "," ++ brk (1,0) ++ do_elem rest in
      hov 1 (str "{" ++ do_elem vl ++ str "}")
    | CVfix (x', defl, x, a, e, _, _) ->
      (* Create a' *)
      let push_ex (id, t) tt =
        Prelude.foralltK t (Term.mkLambda (Names.Name id, t, tt)) in
      let a' = List.fold_right push_ex defl (Prelude.arrowK a) in
      let a' = Prelude.funcK a' in
      (* create the extended environments *)
      let push_tvar tctx (id, t) =
        Environ.push_rel (Names.Name id, None, t) tctx in
      let tctx' = List.fold_left push_tvar tctx defl in
      let defl_len = List.length defl in
      let (ectx', a'') =
        if defl_len > 0 then
          let lift_decl (id, body, t) = (id, body, Term.lift defl_len t) in
          let e_rel_ctx = List.map lift_decl (Environ.rel_context ectx) in
          (Environ.push_rel_context e_rel_ctx Environ.empty_env,
           Term.lift defl_len a')
        else
          (ectx, a') in
      let ectx' = Environ.push_rel (Names.Name x', None, a'') ectx' in
      let ectx' = Environ.push_rel (Names.Name x, None, a) ectx' in
      (* print *)
      let rec do_elem l =
        match l with
        | [] -> str ""
        | (id, t) :: [] ->
          pr_id id ++ str ":" ++ pr_tt_env tctx t
        | (id, t) :: rest ->
          pr_id id ++ str ":" ++ pr_tt_env tctx t ++ str "," ++ brk(1,0) ++
          do_elem rest in
      hov 2 (
        str "fix " ++ pr_id x' ++ str " [" ++ brk(1,0) ++
        hov 0 (do_elem defl) ++ str "]" ++ brk(1,0) ++
        str "(" ++ pr_id x ++ str ":" ++ pr_tt_env tctx' a ++ str ")." ++
        brk(1,0) ++ ewalk tctx' ectx' e)
  in
  vwalk tctx ectx cpsv

let pr_value_tex tctx ectx cpsv =
  let pr_tt_env = Printer.pr_constr_env in
  let pr_id id =
    let id_str = Names.string_of_id  id in
    if (String.get id_str 0) = '_' then
      let next = String.index_from id_str 1 '_' in
      if (String.get id_str 1) = 'k' then
        str "k"
      else
        str (String.sub id_str 1 (next - 1))
    else
      str id_str in
  let rec ewalk tctx ectx e =
    match e with
    | CEapp (v', tl, v) ->
      let rec do_elem l =
        match l with
        | [] -> str ""
        | el :: [] -> pr_tt_env tctx el
        | el :: rest ->
          pr_tt_env tctx el ++ str "," ++ brk(0,0) ++ do_elem rest in
      if tl = [] then
        hov 1 (
          str "\\nappK{" ++ brk(0,0) ++
          vwalk tctx ectx v' ++ str "}{" ++ brk(0,0) ++
          vwalk tctx ectx v ++ str "}")
      else
        hov 1 (
          str "\\nappKt{" ++ brk(0,0) ++
          vwalk tctx ectx v' ++ str "}{" ++ brk(0,0) ++
          do_elem tl ++ str "}{" ++ brk(0,0) ++
          vwalk tctx ectx v ++ str "}")
    | CEval (_, _, _) -> (*CEval (x, v, e) ->*)
      (*
      let v_typ = snd (type_of_value tctx ectx v) in
      let ectx' = Environ.push_rel (Names.Name x, None, v_typ) ectx in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk(1,1) ++
        vwalk tctx ectx v ++ brk(1,0) ++ str "in" ++ brk(1,1) ++
        ewalk tctx ectx' e)
      *)
      str "CEval_NOT_YET"
    | CEunpack (_, _, _, _) -> (*CEunpack (x1, x2, v, e) ->*)
      (*
      let v_typ = snd (type_of_value tctx ectx v) in
      let (b, t) = Prelude.dest_existsK tctx v_typ in
      let ectx', tctx' = Prelude.tctx_push (Names.Name x1, None, b) ectx tctx in
      let a = Term.mkApp (t, [| Term.mkRel 1 |]) in
      let ectx' = Environ.push_rel (Names.Name x2, None, a) ectx' in
      hov 0 (
        str "let " ++ pr_id x1 ++ str ", " ++ pr_id x2 ++ str " = open" ++
        brk(1,1) ++ vwalk tctx ectx v ++ brk(1,0) ++ str "in" ++ brk(1,1) ++
        ewalk tctx' ectx' e)
      *)
      str "CEunpack_NOT_YET"
    | CEsel (x, a, v, v', e) ->
      let v_typ = snd (type_of_value tctx ectx v) in
      let b = snd (Prelude.dest_tupleK tctx v_typ) in
      let v'_typ = snd (type_of_value tctx ectx v') in
      let a' = fst (Prelude.dest_arithmeticK tctx v'_typ) in
      let ba' = Term.mkApp (b, [| a' |]) in
      let ectx' = Environ.push_rel (Names.Name x, None, ba') ectx in
      hov 1 (
        str "\\neselK{" ++ pr_id x ++ str "}{" ++
        pr_tt_env tctx a ++ str "}{" ++ brk(0,0) ++
        vwalk tctx ectx v ++ str "}{" ++ brk(0,0) ++
        vwalk tctx ectx v' ++ str "}{" ++ brk(0,0) ++
        ewalk tctx ectx' e ++ str "}")
    | CEunop (_, _, _, _) -> (*CEunop (x, op, v, e) ->*)
      (*
      let v_typ = snd (type_of_value tctx ectx v) in
      let x_typ =
        match op with
        | Clops.UOnot ->
          Prelude.sboolK
            (Prelude.unop_bool op (Prelude.dest_boolK tctx v_typ)) in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      let op_str = Clops.pretty_unop op in
      hov 0 (
        str "let " ++ pr_id x ++ str " =" ++ brk(1,1) ++
        str op_str ++ hov (String.length op_str) (vwalk tctx ectx v) ++
        brk(1,0) ++ str "in" ++ brk(1,1) ++ ewalk tctx ectx' e)
      *)
      str "CEunop_NOT_YET"
    | CEbinop (x, op, v, v', e) ->
      let v_typ = snd (type_of_value tctx ectx v) in
      let v'_typ = snd (type_of_value tctx ectx v') in
      let x_typ =
        match op with
        | Clops.Blogicop _ ->
          Prelude.sboolK
            (Prelude.binop_bool op
              (Prelude.dest_boolK tctx v_typ) (Prelude.dest_boolK tctx v'_typ))
        | Clops.Barithop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sintK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.snatK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end
        | Clops.Brelop _ ->
          begin
            let dv_typ = Prelude.dest_arithmeticK tctx v_typ in
            let dv'_typ = Prelude.dest_arithmeticK tctx v'_typ in
            match dv_typ, dv'_typ with
            | ((t1, true), (t2, true)) ->
              Prelude.sboolK (Prelude.binop_Z op t1 t2)
            | ((t1, false), (t2, false)) ->
              Prelude.sboolK (Prelude.binop_N op t1 t2)
            | _ -> raise (Prelude.Bad_type v_typ)
          end in
      let ectx' = Environ.push_rel (Names.Name x, None, x_typ) ectx in
      let op_str = Clops.pretty_binop op in
      hov 1 (
        str "\\nebinopK{" ++ pr_id x ++ str "}{" ++
        str op_str ++ str "}{" ++ brk(0,0) ++
        vwalk tctx ectx v ++ str "}{" ++ brk(0,0) ++
        vwalk tctx ectx v' ++ str "}{" ++ brk(0,0) ++
        ewalk tctx ectx' e ++ str "}")
    | CEif (b, a, v, x1, e1, x2, e2) -> 
      let b_tru = Term.mkApp (b, [| Prelude.cterm "true" |]) in
      let ectxbt, tctxbt =
        Prelude.tctx_push (Names.Name x1, None, b_tru) ectx tctx in
      let b_fls = Term.mkApp (b, [| Prelude.cterm "false" |]) in
      let ectxbf, tctxbf =
        Prelude.tctx_push (Names.Name x2, None, b_fls) ectx tctx in
      hov 1 (
        str "\\neifK{" ++ pr_tt_env tctx b ++ str "}{" ++
        pr_tt_env tctx a ++ str "}{" ++ brk(0,0) ++
        vwalk tctx ectx v ++ str "}{" ++
        pr_id x1 ++ str "}{" ++ brk(0,0) ++
        ewalk tctxbt ectxbt e1 ++ str "}{" ++
        pr_id x2 ++ str "}{" ++ brk(0,0) ++
        ewalk tctxbf ectxbf e2 ++ str "}")
  and vwalk tctx ectx v =
    match v with
    | CVvar i ->
      let (var, _, _) = Environ.lookup_rel i ectx in
      hov 0 (str "\\nevarK{" ++ pr_id (Prelude.id_of_name var) ++ str "}")
    | CVnat t ->
      hov 0 (str "\\nenatK{" ++ pr_tt_env tctx t ++ str "}")
    | CVint t ->
      hov 0 (str "\\neintK{" ++ pr_tt_env tctx t ++ str "}")
    | CVbool t ->
      hov 0 (str "\\neboolK{" ++ pr_tt_env tctx t ++ str "}")
    | CVpack (a, v, t) ->
      hov 1 (
        str "\\npackK{" ++ pr_tt_env tctx a ++ str "}{" ++ brk(0,0) ++
        vwalk tctx ectx v ++ str "}{" ++ pr_tt_env tctx t ++ str "}")
    | CVtuple vl ->
      let rec do_elem l =
        match l with
        | [] -> str ""
        | el :: [] -> vwalk tctx ectx el
        | el :: rest ->
          vwalk tctx ectx el ++ str "," ++ brk (0,0) ++ do_elem rest in
      hov 1 (str "\\ntupleK{" ++ brk(0,0) ++ do_elem vl ++ str "}")
    | CVfix (x', defl, x, a, e, _, _) ->
      (* Create a' *)
      let push_ex (id, t) tt =
        Prelude.foralltK t (Term.mkLambda (Names.Name id, t, tt)) in
      let a' = List.fold_right push_ex defl (Prelude.arrowK a) in
      let a' = Prelude.funcK a' in
      (* create the extended environments *)
      let push_tvar tctx (id, t) =
        Environ.push_rel (Names.Name id, None, t) tctx in
      let tctx' = List.fold_left push_tvar tctx defl in
      let defl_len = List.length defl in
      let (ectx', a'') =
        if defl_len > 0 then
          let lift_decl (id, body, t) = (id, body, Term.lift defl_len t) in
          let e_rel_ctx = List.map lift_decl (Environ.rel_context ectx) in
          (Environ.push_rel_context e_rel_ctx Environ.empty_env,
           Term.lift defl_len a')
        else
          (ectx, a') in
      let ectx' = Environ.push_rel (Names.Name x', None, a'') ectx' in
      let ectx' = Environ.push_rel (Names.Name x, None, a) ectx' in
      (* print *)
      let rec do_elem l =
        match l with
        | [] -> str ""
        | (id, t) :: [] ->
          pr_id id ++ str ":" ++ pr_tt_env tctx t
        | (id, t) :: rest ->
          pr_id id ++ str ":" ++ pr_tt_env tctx t ++ str "," ++ brk(0,0) ++
          do_elem rest in
      let x_str = Names.string_of_id x in
      let x'_str = Names.string_of_id x' in
      if List.length defl == 1 then
        hov 1 (
          str "(\\nLambdaK{" ++ pr_id (fst (List.hd defl)) ++ str "}{" ++
          pr_tt_env tctx (snd (List.hd defl)) ++ str "}{" ++
          pr_id x ++ str "}{" ++ pr_tt_env tctx' a ++ str "}{" ++ 
          brk(0,0) ++ ewalk tctx' ectx' e ++ str "})")
      else if (String.sub x'_str 0 3) = "_lk" then
        if (String.sub x_str 0 2) = "_k" then
          hov 1 (
            str "(\\nlambdakK{" ++ brk(0,0) ++ ewalk tctx' ectx' e ++ str "})")
        else
          hov 1 (
            str "(\\nlambdaK{" ++ pr_id x ++ str "}{" ++
            pr_tt_env tctx' a ++ str "}{" ++
            brk(0,0) ++ ewalk tctx' ectx' e ++ str "})")
      else
        hov 1 (
          str "(\\nfixK{"++ pr_id x' ++ str "}{" ++
          hov 1 (do_elem defl) ++ str "}{" ++ pr_id x ++ str "}{" ++
          pr_tt_env tctx' a ++ str "}{" ++
          brk(0,0) ++ ewalk tctx' ectx' e ++ str "})")
  in
  vwalk tctx ectx cpsv

