open Prelude
open Format
open Clops
open Pp

type identifier = Names.identifier

type ('term, 'expr) expr_node =
        | Evar of int
        | Ecnat of 'term
        | Ecint of 'term
        | Ecbool of 'term
        | Elambda of identifier * 'term * 'expr
        | Epoly of identifier * 'term * 'expr
        | Efix of identifier * 'term * 'expr
        | Eapp of 'expr * 'expr
        | Etapp of 'expr * 'term
        | Eif of 'term * 'term * 'expr * identifier * 'expr * identifier * 'expr
        | Etuple of 'expr list
        | Esel of 'term * 'expr * 'expr
        | Epack of 'term * 'expr * 'term
        | Eunpack of 'expr * identifier * identifier * 'expr
        | Eunop of unop * 'expr
        | Ebinop of binop * 'expr * 'expr

type aexpr = (term, aexpr) expr_node * term

let expr x = x

let aevar n at = expr (Evar n, at)
let aecnat t at = expr (Ecnat t, at)
let aecint t at = expr (Ecint t, at)
let aecbool t at = expr (Ecbool t, at)
let aelambda id t e at = expr (Elambda (id, t, e), at) 
let aepoly id t e at = expr (Epoly (id, t, e), at) 
let aefix id t f at = expr (Efix (id, t, f), at)
let aeapp e1 e2 at = expr (Eapp (e1, e2), at) 
let aetapp e1 t at = expr (Etapp (e1, t), at)
let aeif b a e x1 e1 x2 e2 at = expr (Eif (b, a, e, x1, e1, x2, e2), at) 
let aetupple el at = expr (Etuple el, at)
let aesel a e1 e2 at = expr (Esel (a, e1, e2), at)
let aepack a e t at = expr (Epack (a, e, t), at)
let aeunpack e1 x1 x2 e2 at = expr (Eunpack (e1, x1, x2, e2), at)
let aeunop op e1 at = expr (Eunop (op, e1), at) 
let aebinop op e1 e2 at = expr (Ebinop (op, e1, e2), at) 

let pr_aexpr tctx ectx ae =
  let pr_tt = Printer.pr_constr in
  let pr_tt_env = Printer.pr_constr_env in
  let pr_id id = str (Names.string_of_id id) in
  let rec walk tctx ectx ea =
    let (e, _) = ea in
    match e with
      | Evar n ->
        let (var, _, _) = Environ.lookup_rel n ectx in
        hov 0 (str (string_of_name var))
      | Ecnat t ->
        hov 0 (str "cnat [" ++ pr_tt t ++ str "]")
      | Ecint t ->
        hov 0 (str "cint [" ++ pr_tt t ++ str "]")
      | Ecbool t ->
        hov 0 (str "cbool [" ++ pr_tt t ++ str "]")
      | Elambda (id, t, e) ->
        let ectx = Environ.push_rel (Names.Name id, None, t) ectx in
        hov 2 (
          str "lambda " ++ pr_id id ++ str ":" ++ pr_tt_env tctx t ++ 
          str "." ++ brk (1,0) ++ walk tctx ectx e)
      | Epoly (id, t, e) ->
        let ectx, tctx' = tctx_push (Names.Name id, None, t) ectx tctx in
        hov 2 (
          str "poly " ++ pr_id id ++ str ":" ++ pr_tt_env tctx t ++
          str "." ++ brk (1,0) ++ walk tctx' ectx e)
      | Eunpack (e1, x1, x2, e2) ->
        let (b, t') = dest_exists tctx (snd e1) in
        let ectx', tctx' = tctx_push (Names.Name x1, None, b) ectx tctx in
        let t'X = Term.mkApp (t', [| Term.mkRel 1 |]) in
        let ectx' = Environ.push_rel (Names.Name x2, None, t'X) ectx' in
        hov 0 (
          str "unpack" ++ brk (1,2) ++ walk tctx ectx e1 ++ brk (1,0) ++
          str "as (" ++ pr_id x1 ++ str ", " ++ pr_id x2 ++ str ") in" ++
          brk (1,2) ++ walk tctx' ectx' e2)
      | Efix (id, t, e) ->
        let ectx' = Environ.push_rel (Names.Name id, None, t) ectx in
        hov 2 (
          str "fix " ++ pr_id id ++ str ":" ++ pr_tt_env tctx t ++
          str "." ++ brk (1,0) ++ walk tctx ectx' e)
      | Eapp (e1, e2) ->
        hov 0 (
          hov 1 (str "(" ++ walk tctx ectx e1 ++ str ")") ++ brk (1,0) ++
          hov 1 (str "(" ++ walk tctx ectx e2 ++ str ")"))
      | Etapp (e, t) ->
        hov 0 (
          hov 1 (str "(" ++ walk tctx ectx e ++ str ")") ++ brk (1,0) ++
          hov 1 (str "[" ++ pr_tt_env tctx t ++ str "]"))
      | Eif (b, a, e, x1, e1, x2, e2) ->
        let b_tru = Term.mkApp (b, [| cterm "true" |]) in
        let b_fls = Term.mkApp (b, [| cterm "false" |]) in
        let ectxbt, tctxbt =
          tctx_push (Names.Name x1, None, b_tru) ectx tctx in
        let ectxbf, tctxbf =
          tctx_push (Names.Name x2, None, b_fls) ectx tctx in
        hov 2 (
          str "if [" ++ pr_tt_env tctx b ++ str ", " ++ pr_tt_env tctx a ++
          str "]" ++ brk (1,0) ++
          str "(" ++ walk tctx ectx e ++ str "," ++ brk (1,0) ++
          pr_id x1 ++ str "." ++ walk tctxbt ectxbt e1 ++ str "," ++
          brk (1,0) ++ pr_id x2 ++ str "." ++ walk tctxbf ectxbf e2 ++ str ")")
      | Etuple el ->
        let rec do_elem l =
          match l with
          | [] -> str ""
          | el :: [] -> walk tctx ectx el
          | el :: rest ->
            walk tctx ectx el ++ str "," ++ brk (1,0) ++ do_elem rest in
        hov 1 (str "{" ++ do_elem el ++ str "}")
      | Esel (a, e1, e2) ->
        hov 2 (
          str "sel [" ++ pr_tt_env tctx a ++ str "]" ++ brk (1,0) ++
          str "(" ++ walk tctx ectx e1 ++ str "," ++ brk (1,0) ++
          walk tctx ectx e2 ++ str ")")
      | Epack (a, e, t) ->
        hov 0 (
          str "pack" ++ brk (1,2) ++
          str "(" ++ pr_tt_env tctx a ++ str "," ++ brk (1,2) ++
          walk tctx ectx e ++ str ")" ++ brk (1,0) ++ str "as" ++ brk (1,2) ++
          pr_tt_env tctx t)
      | Eunop (op, e) ->
        hov 0 (str (pretty_unop op) ++ str "(" ++ walk tctx ectx e ++ str ")")
      | Ebinop (op, e1, e2) ->
        hov 0 (
          walk tctx ectx e1 ++ str " " ++ str (pretty_binop op) ++ brk (1,0) ++
          walk tctx ectx e2)
    in
      walk tctx ectx ae

