open Sys
open Names

open Pp

type verbose = Vquiet | Vnormal | Vverbose
let flagVerbose = ref Vnormal

let process_file fname =
  if !flagVerbose = Vverbose || !flagVerbose = Vnormal then
    msgnl (str ("Processing file: " ^ fname));
  Vernac.load_vernac (!flagVerbose = Vverbose) fname;
  if !flagVerbose = Vverbose || !flagVerbose = Vnormal then
    msgnl (str ("End of file: " ^ fname))

let set_terminal_size maxcol =
  Format.set_max_indent (maxcol - 20);
  Format.set_margin maxcol

let set_formater_size fmt =
  Format.pp_set_max_indent fmt 150;
  Format.pp_set_margin fmt 170

(* calculate the line and column. taken from Coq sources *)
let calc_linepos fname offset =
  try
    let fstat = Unix.stat fname in
    if fstat.Unix.st_kind == Unix.S_REG then
      let stream = open_in fname in
      let rec line_of_pos lin bol cnt =
        if cnt < offset then
          if input_char stream == '\n' then
            line_of_pos (lin + 1) (cnt + 1) (cnt + 1)
          else
            line_of_pos lin bol (cnt + 1)
        else (lin, offset - bol) in
      let (line, bol) = line_of_pos 1 0 0 in
      close_in stream;
      str "line " ++ int line ++ str ", character " ++ int bol
    else
      str "unknown position"
  with _ -> str "unknown position"

(* COQ toplevel *)
let toplevel_default_name = Names.make_dirpath [Names.id_of_string "Top"]
let toplevel_name = ref (Some toplevel_default_name)

(* COQ inputstate *)
let inputstate = ref None
let inputstate () =
  match !inputstate with
    | Some "" -> ()
    | Some s -> States.intern_state s
    | None -> States.intern_state "initial.coq"

(* COQ vm_opt *)
let boxed_val = ref false
let boxed_def = ref false
let use_vm = ref false

let set_vm_opt () =
  Vm.set_transp_values (not !boxed_val);
  Options.set_boxed_definitions !boxed_def;
  Vconv.set_use_vm !use_vm

(* Objects to link, copied from coqmktop.ml *)

let spaces = Str.regexp "[ \t\n]+"
let split_list l = Str.split spaces l
let core_objs = split_list Tolink.core_objs

(* Transforms a file name in the corresponding Caml module name. *)
let rem_ext_regexpr = Str.regexp "\\(.*\\)\\.\\(cm..?\\|ml\\)"

let module_of_file name =
  let s = Str.replace_first rem_ext_regexpr "\\1" (Filename.basename name) in
  String.capitalize s

(* Back to earth... *)

external columns  : unit -> int
    = "c_columns"
external rows  : unit -> int
    = "c_rows"

let main =
  (* from booting: add the pre-linked modules *)
  let modules = List.map module_of_file core_objs in
  List.iter Mltop.add_known_module modules;
  (* from Coqtop.start *)
  Options.batch_mode := true;
  Options.make_silent true;
  Lib.init ();
  Coqinit.init_load_path ();
  inputstate ();
  set_vm_opt ();
  Global.set_engagement Declarations.ImpredicativeSet;
  if Global.env_is_empty() then
    Util.option_iter Declaremods.start_library !toplevel_name;
  Coqinit.init_library_roots ();
  (* let init_state = States.freeze() in *)
  if ( (Array.length Sys.argv) == 1 ) then
    begin msgnl (str "No file!"); exit 1 end;
  let filename = Array.get Sys.argv 1 in
  Pp_control.set_gp Format.std_formatter Pp_control.deep_gp;
  set_terminal_size (columns() * 95 / 100);
  try
    (* States.unfreeze init_state; *)
    process_file "Prelude.v";
    Prelude.init ();
    let instream = open_in filename in
    let strm = Stream.of_channel instream in
    Parser.init ();
    (* parsing and type checking *)
    let expr_f = Parser.parse strm in
    let coq_env = Global.env () in
    let aexpr = expr_f Environ.empty_env coq_env in
    let aexpr_typ = snd aexpr in
    Format.printf "Computational Expression :\n@?";
    Pp.ppnl_with Format.std_formatter
      (Expr.pr_aexpr coq_env Environ.empty_env aexpr);
    Format.printf "@?\nComputational Type :\n@?";
    Pp.ppnl_with Format.std_formatter (Printer.pr_constr aexpr_typ);
    (* CPS conversion *)
    let cpsexpr = Cpsexpr.aexpr_to_cps aexpr in
    Format.printf "@?\nCPS Expression :\n@?";
    Pp.ppnl_with Format.std_formatter
      (Cpsexpr.pr_value coq_env Environ.empty_env cpsexpr);
    (* CPS type checking *)
    let (cpsexpr, cpsexpr_typ) =
      Cpsexpr.type_of_value coq_env Environ.empty_env cpsexpr in
    Format.printf "@?\nCPS Type:\n@?";
    Pp.ppnl_with Format.std_formatter (Printer.pr_constr cpsexpr_typ);
    (* CPS sanity check *)
    if Prelude.eq_OmegaK coq_env
      (Prelude.cps_correctness aexpr_typ) cpsexpr_typ then
      Format.printf "@?\nCPS sanity check ok\n@?"
    else
      Format.printf "@?\nCPS sanity check FAILED!\n@?";
    let ofilename = "cps.tex" in
    let ofile = open_out ofilename in
    let cps_tex = Format.formatter_of_out_channel ofile in
    Pp.ppnl_with cps_tex
     (Cpsexpr.pr_value_tex coq_env Environ.empty_env cpsexpr);
    close_out ofile;
    Format.printf "@?\nCPS term written in %s\n@?" ofilename;
    (* Closure conversion *)
    let closval = Closexpr.cpsval_to_closval cpsexpr in
    Format.printf "@?\nClosure Expression :\n@?";
    Pp.ppnl_with Format.std_formatter
      (Closexpr.pr_value coq_env Environ.empty_env closval);
    (* closure type checking *)
    let closexpr_typ =
      Closexpr.type_of_value coq_env Environ.empty_env closval in
    Format.printf "@?\nClosure Type:\n@?";
    Pp.ppnl_with Format.std_formatter (Printer.pr_constr closexpr_typ);
    let closure_correct =
      Term.mkApp
        (Prelude.cterm "to_closure", [| cpsexpr_typ; Prelude.tupleK [] |]) in
    if Prelude.eq_OmegaK coq_env closexpr_typ closure_correct then
      Format.printf "@?\nClosure sanity check ok\n@?"
    else
      Format.printf "@?\nClosure sanity check FAILED!\n@?";
  with
    | Stdpp.Exc_located ((bp, _), Parser.ParseError cmds) ->
      let pos = calc_linepos filename bp.Lexing.pos_cnum in
      msg (str "Error in " ++ pos ++ str " in ");
      Pp.ppnl_with Format.std_formatter cmds
    | Stdpp.Exc_located ((bp, _), e) ->
      msgnl (str "Error in " ++ calc_linepos filename bp.Lexing.pos_cnum);
      msgnl (Toplevel.print_toplevel_error e)
    | Prelude.Type_Error cmds ->
      msg (str "Internal Type Error in ");
      Pp.ppnl_with Format.std_formatter cmds
    | e ->
      msgnl (Toplevel.print_toplevel_error e)

