open Expr
open Prelude
open Clops
open Pp

exception ParseError of std_ppcmds

module Gram = Pcoq.Gram

module CL :
  sig
    val expr : (Environ.env -> Environ.env -> aexpr) Gram.Entry.e
    val start : (Environ.env -> Environ.env -> aexpr) Gram.Entry.e
    val term : (Environ.env -> term) Gram.Entry.e
  end
=
  struct
    let start = Gram.Entry.create "start"
    let expr = Gram.Entry.create "expr"
    let term = Gram.Entry.create "term"
  end

let err loc e = Stdpp.raise_with_loc loc e

(* Typing for binary operators *)
let do_binop loc op e1 e2 ectx tctx =
  let (e1e, e1t) = e1 ectx tctx in
  let (e2e, e2t) = e2 ectx tctx in
  try
    match op with
    | Blogicop _ ->
      let t1 = dest_bool tctx e1t in
      let t2 = dest_bool tctx e2t in
        aebinop op (e1e, e1t) (e2e, e2t) (sbool (binop_bool op t1 t2))
    | Barithop _ ->
      begin
        match (dest_arithmetic tctx e1t, dest_arithmetic tctx e2t) with
        | ((t1, true), (t2, true)) ->
          aebinop op (e1e, e1t) (e2e, e2t) (sint (binop_Z op t1 t2))
        | ((t1, false), (t2, false)) ->
          aebinop op (e1e, e1t) (e2e, e2t) (snat (binop_N op t1 t2))
        | _ ->
          raise (Bad_type e1t)
      end
    | Brelop _ ->
      begin
        match (dest_arithmetic tctx e1t, dest_arithmetic tctx e2t) with
        | ((t1, true), (t2, true)) ->
          aebinop op (e1e, e1t) (e2e, e2t) (sbool (binop_Z op t1 t2))
        | ((t1, false), (t2, false)) ->
          aebinop op (e1e, e1t) (e2e, e2t) (sbool (binop_N op t1 t2))
        | _ ->
          raise (Bad_type e1t)
      end
  with Bad_type _ ->
    err loc (ParseError (
      v 0 (str "binary operator " ++ str (pretty_binop op) ++ brk (0,0) ++
        hv 0 (
          str "Expression 1 :" ++ brk (1, 2) ++
          hov 0 (pr_aexpr tctx ectx (e1e, e1t)) ++ brk (1,0) ++
          str "has type :" ++ brk (1, 2) ++
          hov 0 (Printer.pr_constr_env tctx e1t)) ++ brk (0,0) ++
        hv 0 (
          str "Expression 2 :" ++ brk (1, 2) ++
          hov 0 (pr_aexpr tctx ectx (e2e, e1t)) ++ brk (1,0) ++
          str "has type :" ++ brk (1, 2) ++
          hov 0 (Printer.pr_constr_env tctx e2t)))))

(* Typing for unary operators *)
let do_unop loc op e ectx tctx =
  let (ee, et) = e ectx tctx in
  try
    match op with
    | UOnot ->
      let t = dest_bool tctx et in
      aeunop UOnot (ee, et) (sbool (unop_bool op t))
  with Bad_type _ ->
    err loc (ParseError (
      v 0 (str "unary operator " ++ str (pretty_unop op) ++ brk (0,0) ++
        hv 0 (
          str "Expression :" ++ brk (1, 2) ++
          hov 0 (pr_aexpr tctx ectx (ee, et)) ++ brk (1,0) ++
          str "has type :" ++ brk (1, 2) ++
          hov 0 (Printer.pr_constr_env tctx et)))))

let init () =
  let constr_kw =
    [ "lambda"; "poly"; "if"; "cnat"; "cint"; "cbool"; "sel";
      "pack"; "unpack"; "as"; "in"; "fix"; "%%%" ] in
  List.iter (fun s -> Lexer.add_token("",s)) constr_kw;

open CL


GEXTEND Gram
  GLOBAL: start;
  expr:
    [ RIGHTA
      [ f = func -> f
      | "fix"; id = ident; ":"; t = term; "."; f = func ->
          fun ectx tctx ->
          let typ = t tctx in
          let t_typ = term_type tctx typ in
          if term_equal_env tctx t_typ (omega ()) then
            let ectx' = Environ.push_rel (Names.Name id, None, typ) ectx in
            let af = f ectx' tctx in
            let (_, aft) = af in
            if term_equal_env tctx typ aft then
              aefix id typ af typ
            else
              err _loc (ParseError (
                v 0 (str "body of fix" ++ brk (0,0) ++
                  hv 0 (
                    str "Type is :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx aft) ++ brk (1,0) ++
                    str "but should be :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx typ)))))
          else
            err _loc (ParseError (
              v 0 (str "fix return type" ++ brk (0,0) ++
                hv 0 (
                  str "Type of :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx typ) ++ brk (1,0) ++
                  str "should be Omega, but is :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx t_typ)))))
      | "unpack"; e1 = expr; "as";
        "("; x1 = ident; ","; x2 = ident; ")"; "in"; e2 = expr ->
          fun ectx tctx ->
          begin
            let ae1 = e1 ectx tctx in
            let (_, e1t) = ae1 in
            try
              let (b, t') = dest_exists tctx e1t in
              let ectx', tctx' = tctx_push (Names.Name x1, None, b) ectx tctx in
              let t'X = Term.mkApp (t', [| Term.mkRel 1 |]) in
              let ectx' = Environ.push_rel (Names.Name x2, None, t'X) ectx' in
              let ae2 = e2 ectx' tctx' in
              let (_, e2t) = ae2 in
              if Term.noccurn 1 e2t then
                aeunpack ae1 x1 x2 ae2 (Term.lift (-1) e2t)
              else
                err _loc (ParseError (
                  v 0 (str "unpack" ++ brk (0,0) ++
                    hv 0 (
                      str "Expression :" ++ brk (1, 2) ++
                      hov 0 (pr_aexpr tctx' ectx' ae2) ++ brk (1,0) ++
                      str "exposes local variables (" ++
                      str (Names.string_of_id x1) ++ str " or " ++
                      str (Names.string_of_id x2) ++ str ")"))))
            with Bad_type _ ->
              err _loc (ParseError (
                v 0 (str "unpack argument" ++ brk (0,0) ++
                  hv 0 (
                    str "Type of :" ++ brk (1, 2) ++
                    hov 0 (pr_aexpr tctx ectx ae1) ++ brk (1,0) ++
                    str "should be existsT, but is :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx e1t)))))
          end
      ]
    | NONA
      [ "if"; "["; b = term; ","; a = term; "]"; "("; e = expr; ",";
          x1 = ident; "."; e1 = expr; ","; x2 = ident; "."; e2 = expr; ")" ->
          fun ectx tctx ->
          begin
            let bt = b tctx in
            let btos = Term.mkArrow (cterm "bool") Term.mkProp in
            let btt = term_type tctx bt in
            if term_equal_env tctx btt btos then
              let (ee, et) = e ectx tctx in
              try
                let e_arg = dest_bool tctx et in
                let at = a tctx in
                let att = term_type tctx at in
                let bt'e_arg = Term.mkApp (bt, [| e_arg |]) in
                if term_equal_env tctx att bt'e_arg then
                  let bt_tru = Term.mkApp (bt, [| cterm "true" |]) in
                  let ectxbt, tctxbt =
                    tctx_push (Names.Name x1, None, bt_tru) ectx tctx in
                  let (e1e, e1t) = e1 ectxbt tctxbt in
                  let bt_fls = Term.mkApp (bt, [| cterm "false" |]) in
                  let ectxbf, tctxbf =
                    tctx_push (Names.Name x2, None, bt_fls) ectx tctx in
                  let (e2e, e2t) = e2 ectxbf tctxbf in
                  if not (Term.noccurn 1 e1t) then
                    err _loc (ParseError (
                      v 0 (str "then clause" ++ brk (0,0) ++
                        hv 0 (
                          str "Expression :" ++ brk (1, 2) ++
                          hov 0 (pr_aexpr tctxbt ectxbt (e1e, e1t)) ++
                          brk (1,0) ++ str "exposes the proof (" ++
                          Nameops.pr_id x1 ++ str ")!"))))
                  else if not (Term.noccurn 1 e2t) then
                    err _loc (ParseError (
                      v 0 (str "else clause" ++ brk (0,0) ++
                        hv 0 (
                          str "Expression :" ++ brk (1, 2) ++
                          hov 0 (pr_aexpr tctxbf ectxbf (e2e, e2t)) ++
                          brk (1,0) ++ str "exposes the proof (" ++
                          Nameops.pr_id x2 ++ str ")!"))))
                  else if term_equal_env tctxbt e1t e2t then
                    let ebt = Term.lift (-1) e1t in
                    aeif bt at (ee, et) x1 (e1e, e1t) x2 (e2e, e2t) ebt
                  else
                    err _loc (ParseError (
                      v 0 (str "if clauses" ++ brk (0,0) ++
                        hv 0 (
                          str "then expression :" ++ brk (1, 2) ++
                          hov 0 (pr_aexpr tctxbt ectxbt (e1e, e1t)) ++
                          brk (1,0) ++
                          str "has type :" ++ brk (1, 2) ++
                          hov 0 (Printer.pr_constr_env tctxbt e1t)) ++
                          brk (0,0) ++
                        hv 0 (
                          str "else expression :" ++ brk (1, 2) ++
                          hov 0 (pr_aexpr tctxbf ectxbf (e2e, e1t)) ++
                          brk (1,0) ++
                          str "has type :" ++ brk (1, 2) ++
                          hov 0 (Printer.pr_constr_env tctxbf e2t)))))
                else
                  err _loc (ParseError (
                    v 0 (str "second if type argument" ++ brk (0,0) ++
                      hv 0 (
                        str "Type of :" ++ brk (1, 2) ++
                        hov 0 (Printer.pr_constr_env tctx at) ++ brk (1,0) ++
                        str "is :" ++ brk (1, 2) ++
                        hov 0 (Printer.pr_constr_env tctx att) ++ brk (1,0) ++
                        str "but should be:" ++ brk (1, 2) ++
                        hov 0 (Printer.pr_constr_env tctx bt'e_arg)))))
              with Bad_type _ ->
                err _loc (ParseError (
                  v 0 (str "if condition expression" ++ brk (0,0) ++
                    hv 0 (
                      str "Type of :" ++ brk (1, 2) ++
                      hov 0 (pr_aexpr tctx ectx (ee, et)) ++ brk (1,0) ++
                      str "should be bool, but is :" ++ brk (1, 2) ++
                      hov 0 (Printer.pr_constr_env tctx et)))))
            else
              err _loc (ParseError (
                v 0 (str "first if type argument" ++ brk (0,0) ++
                  hv 0 (
                    str "Type of :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx bt) ++ brk (1,0) ++
                    str "should be Bool -> Prop but is:" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx btt)))))
          end
      | "sel"; "["; t = term; "]"; "("; e1 = expr; ","; e2 = expr; ")" ->
          fun ectx tctx ->
          begin
            let (e1e, e1t) = e1 ectx tctx in
            let (e2e, e2t) = e2 ectx tctx in
            let tt = t tctx in
            try
              let (n, f) = dest_tuple tctx e1t in
              try
                let (i, is_int) = dest_arithmetic tctx e2t in
                if not is_int then
                  let i_lt_n = Prelude.lt_prop_t i n in
                  if term_equal_env tctx i_lt_n tt then
                    aesel
                      tt (e1e, e1t) (e2e, e2t) (Term.mkApp (f, [| i |]))
                  else (* FIXME *)
                    err _loc (ParseError (
                      v 0 (str "select proof" ++ brk (0,0) ++
                        hv 0 (
                          str "Type of :" ++ brk (1, 2) ++
                          hov 0 (Printer.pr_constr_env tctx tt) ++ brk (1,0) ++
                          str "is :" ++ brk (1, 2) ++
                          hov 0 (Printer.pr_constr_env tctx tt) ++
                          brk (1,0) ++
                          str "but should be:" ++ brk (1, 2) ++
                          hov 0 (Printer.pr_constr_env tctx i_lt_n)))))
                else
                  raise (Bad_type e2t)
              with Bad_type _ ->
                err _loc (ParseError (
                  v 0 (str "select index" ++ brk (0,0) ++
                    hv 0 (
                      str "Type of :" ++ brk (1, 2) ++
                      hov 0 (pr_aexpr tctx ectx (e2e, e2t)) ++ brk (1,0) ++
                      str "should be in snat, but is :" ++ brk (1, 2) ++
                      hov 0 (Printer.pr_constr_env tctx e2t)))))
            with Bad_type _ ->
              err _loc (ParseError (
                v 0 (str "select argument" ++ brk (0,0) ++
                  hv 0 (
                    str "Type of :" ++ brk (1, 2) ++
                    hov 0 (pr_aexpr tctx ectx (e1e, e1t)) ++ brk (1,0) ++
                    str "should be tuple, but is :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx e1t)))))
          end
      ]
    | LEFTA
      [ e1 = expr; "&"; e2 = expr -> do_binop _loc (Blogicop BOand) e1 e2 ]
    | LEFTA
      [ e1 = expr; "|"; e2 = expr -> do_binop _loc (Blogicop BOor) e1 e2 ]
    | NONA
      [ "!"; e = expr -> do_unop _loc UOnot e ]
    | NONA
      [ e1 = expr; "=="; e2 = expr -> do_binop _loc (Brelop BOeq) e1 e2
      | e1 = expr; "!="; e2 = expr -> do_binop _loc (Brelop BOne) e1 e2
      | e1 = expr; "<";  e2 = expr -> do_binop _loc (Brelop BOlt) e1 e2
      | e1 = expr; ">";  e2 = expr -> do_binop _loc (Brelop BOgt) e1 e2
      | e1 = expr; "<="; e2 = expr -> do_binop _loc (Brelop BOle) e1 e2
      | e1 = expr; ">="; e2 = expr -> do_binop _loc (Brelop BOge) e1 e2
      ]
    | LEFTA
      [ e1 = expr; "+"; e2 = expr -> do_binop _loc (Barithop BOplus) e1 e2
      | e1 = expr; "-"; e2 = expr -> do_binop _loc (Barithop BOminus) e1 e2
      ]
    | LEFTA
      [ e1 = expr; "*"; e2 = expr -> do_binop _loc (Barithop BOtimes) e1 e2
      | e1 = expr; "/"; e2 = expr -> do_binop _loc (Barithop BOdiv) e1 e2
      | e1 = expr; "%"; e2 = expr -> do_binop _loc (Barithop BOmod) e1 e2
      ]
    | LEFTA
      [ e1 = expr; e2 = expr ->
          fun ectx tctx ->
          let e1e, e1t = e1 ectx tctx
          and e2e, e2t = e2 ectx tctx in
          begin
          try
            let (e1t1, e1t2) = dest_arrow tctx e1t in
            if term_equal_env tctx e1t1 e2t then
              aeapp (e1e, e1t) (e2e, e2t) e1t2
            else
              err _loc (ParseError (
                v 0 (str "application" ++ brk (0,0) ++
                  hv 0 (
                    str "Type of :" ++ brk (1, 2) ++
                    hov 0 (pr_aexpr tctx ectx (e2e, e2t)) ++ brk (1,0) ++
                    str "is :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx e2t) ++ brk (1,0) ++
                    str "but should be:" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx e1t1)))))
          with Bad_type _ ->
            err _loc (ParseError (
              v 0 (str "application" ++ brk (0,0) ++
                hv 0 (
                  str "Type of :" ++ brk (1, 2) ++
                  hov 0 (pr_aexpr tctx ectx (e1e, e1t)) ++ brk (1,0) ++
                  str "should be arrow, but is :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx e1t)))))
          end
      | e = expr; "{"; a = term; "}" ->
          fun ectx tctx ->
          let ee, et = e ectx tctx
          and atyp = a tctx in
          begin
          try
            let (a', t) = dest_forall tctx et in
            let t_atyp = term_type tctx atyp in
            if term_equal_env tctx a' t_atyp then
              aetapp (ee, et) atyp (Term.mkApp (t, [| atyp |]))
            else
              err _loc (ParseError (
                v 0 (str "type application" ++ brk (0,0) ++
                  hv 0 (
                    str "Type of :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx atyp) ++ brk (1,0) ++
                    str "is :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx t_atyp) ++ brk (1,0) ++
                    str "but should be:" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx a')))))
          with Bad_type _ ->
            err _loc (ParseError (
              v 0 (str "type application" ++ brk (0,0) ++
                hv 0 (
                  str "Type of :" ++ brk (1, 2) ++
                  hov 0 (pr_aexpr tctx ectx (ee, et)) ++ brk (1,0) ++
                  str "should be poly, but is :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx et)))))
          end
      ]
    | NONA
      [ "cint"; "["; i = integer; "]" ->
          fun _ _ -> aecint i (sint i)
      | "cnat"; "["; n = natural; "]" ->
          fun _ _ -> aecnat n (snat n)
      | "cbool"; "["; b = boolean; "]" ->
          fun _ _ ->
          let b_typ = term_type (Global.env ()) b in
          if term_equal_env (Global.env ()) b_typ (cterm "bool") then
            aecbool b (sbool b)
          else
            err _loc (ParseError (
              v 0 (str "boolean constant" ++ brk (0,0) ++
                hv 0 (
                  str "Term :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr b) ++ brk (1,0) ++
                  str "is not a valid boolean constant"))))
      | "pack"; "("; a = term; ","; e = expr; ")"; "as"; t = term ->
          fun ectx tctx ->
          begin
            let typ = t tctx in
            try
              let (b, t') = dest_exists tctx typ in
              let atyp = a tctx in
              let t_atyp = term_type tctx atyp in
              if term_equal_env tctx t_atyp b then
                let (ee, et) = e ectx tctx in
                let t'a = Term.mkApp (t', [| atyp |]) in
                if term_equal_env tctx et t'a then
                  aepack atyp (ee, et) typ typ
                else
                  err _loc (ParseError (
                    v 0 (str "pack expression" ++ brk (0,0) ++
                      hv 0 (
                        str "Type of :" ++ brk (1, 2) ++
                        hov 0 (pr_aexpr tctx ectx (ee, et)) ++ brk (1,0) ++
                        str "is :" ++ brk (1, 2) ++
                        hov 0 (Printer.pr_constr_env tctx et) ++ brk (1,0) ++
                        str "but should be:" ++ brk (1, 2) ++
                        hov 0 (Printer.pr_constr_env tctx t'a)))))
              else
                err _loc (ParseError (
                  v 0 (str "pack type argument" ++ brk (0,0) ++
                    hv 0 (
                      str "Type of :" ++ brk (1, 2) ++
                      hov 0 (Printer.pr_constr_env tctx atyp) ++ brk (1,0) ++
                      str "is :" ++ brk (1, 2) ++
                      hov 0 (Printer.pr_constr_env tctx t_atyp) ++ brk (1,0) ++
                      str "but should be:" ++ brk (1, 2) ++
                      hov 0 (Printer.pr_constr_env tctx b)))))
            with Bad_type _ ->
              err _loc (ParseError (
                v 0 (str "pack result type" ++ brk (0,0) ++
                  hv 0 (
                    str "Type of :" ++ brk (1, 2) ++
                    hov 0 (Printer.pr_constr_env tctx typ) ++ brk (1,0) ++
                    str "should be existsT"))))
            end
      | "{"; el = LIST1 expr SEP ","; "}" ->
          fun ectx tctx ->
          begin
            let rec walk l =
              match l with
              | [] -> ([], [])
              | hd::tl ->
                let ae = hd ectx tctx in
                let (tl1, tl2) = walk tl in
                (ae::tl1, (snd ae)::tl2) in
            let (elist, tlist) = walk el in
              aetupple elist (tuple tlist)
          end
      ]
    | NONA
      [ c = Pcoq.Constr.constr -> (* identifier! *)
        fun ectx _ ->
        begin
        try
          let rcon = Constrintern.interp_constr Evd.empty ectx c in
          let index = Term.destRel rcon in
          let (_, _, t) = Environ.lookup_rel index ectx in
            aevar index t
        with
          | _ ->
            err _loc (ParseError (
              v 0 (str "identifier" ++ brk (0,0) ++
                Ppconstr.pr_constr_expr c ++
                str " is invalid or was not found in the context")))
        end
      | "("; e = expr; ")" -> e
      ]
    ];
  func:
    [ RIGHTA
      [ "lambda"; id = ident; ":"; t = term; "."; e = expr ->
          fun ectx tctx ->
          let typ = t tctx in
          let t_typ = term_type tctx typ in
          if term_equal_env tctx t_typ (omega ()) then
            let ectx' = Environ.push_rel (Names.Name id, None, typ) ectx in
            let ae = e ectx' tctx in
            let (_, et) = ae in
              aelambda id typ ae (arrow typ et)
          else
            err _loc (ParseError (
              v 0 (str "lambda" ++ brk (0,0) ++
                hv 0 (
                  str "Type of :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx typ) ++ brk (1,0) ++
                  str "should be Omega, but is :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx t_typ)))))
      | "poly"; id = ident; ":"; t = term; "."; f = func ->
          fun ectx tctx ->
          let typ = t tctx in
          let t_typ = term_type tctx typ in
          if Term.isSort (whnf_env tctx t_typ) then
            let ectx, tctx' = tctx_push (Names.Name id, None, typ) ectx tctx in
            let af = f ectx tctx' in
            let (_, ft) = af in
              aepoly id typ af
               (forallt typ (Term.mkLambda (Names.Name id, typ, ft)))
          else
            err _loc (ParseError (
              v 0 (str "poly" ++ brk (0,0) ++
                hv 0 (
                  str "Type of :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx typ) ++ brk (1,0) ++
                  str "should be Sort, but is :" ++ brk (1, 2) ++
                  hov 0 (Printer.pr_constr_env tctx t_typ)))))
      ]
    ];
  term:
    [ [ t = Pcoq.Constr.lconstr ->
        fun tctx ->
          Constrintern.interp_constr Evd.empty tctx t ] ];
  ident:
    [ [ id = Pcoq.Prim.ident -> id ] ];
  natural:
    [ [ n = Pcoq.Prim.natural -> int_to_N n ] ];
  integer:
    [ [ n = Pcoq.Prim.integer -> int_to_Z n ] ];
  boolean:
    [ [ b = Pcoq.Constr.constr  ->
          let t = Constrintern.interp_constr Evd.empty (Global.env ()) b in
            t ] ];
  start:
    [ [ d = LIST0 Pcoq.main_entry; "%%%"; e = expr ->
          try
            let do_vernac el =
              match el with
              | Some v -> Vernac.vernac_com Vernacentries.interp v
              | None -> () in
            List.iter do_vernac d;
            e
          with Vernac.DuringCommandInterp (_, e) -> raise e
      ] ];
END

let parse strm =
  Gram.Entry.parse (Pcoq.eoi_entry start) (Gram.parsable strm)

