
type term = Term.constr

open Pp
exception Bad_type of term
exception Internal_Error of string
exception Type_Error of Pp.std_ppcmds

(* Map from string (COQ constructor) to Term.constr *)
module CTmap = Map.Make(String)
let terms = ref CTmap.empty

(* These terms should exist in Prelude.v or the COQ library.
   Upon calling init() the Map above will be filled with these *)
let coq_terms = [
  "list"; "cons"; "nil"; (* list *)
  "bool"; "true"; "false"; (* bool *)
  "bool_not"; "bool_or"; "bool_and"; (* bool ops *)
  "Z_plus"; "Z_minus"; "Z_times"; "Z_div"; "Z_mod"; (* Z (int) ops *)
  "Z_eq"; "Z_ne"; "Z_lt"; "Z_le"; "Z_gt"; "Z_ge";
  "N_plus"; "N_minus"; "N_times"; "N_div"; "N_mod"; "Nsucc"; (* N (nat) ops *)
  "N_eq"; "N_ne"; "N_lt"; "N_le"; "N_gt"; "N_ge"; "N_lt_prop";
  "snat"; "sint"; "sbool"; (* Omega *)
  "arrow"; "tuple"; "forallT"; "existsT"; "Omega";
  "snatK"; "sintK"; "sboolK"; (* OmegaK *)
  "arrowK"; "tupleK"; "forallTK"; "existsTK"; "funcK"; "OmegaK";
  "to_OmegaK"; (* Omega -> OmegaK *)
  "cps_correctness"; (* CPS expected type. Omega -> OmegaK *)
  "list_to_lambda"; (* convert list to function abstraction *)
  "to_closure"; (* closure transformation *)
]

(* The actual function that does the conversion.
   Should be called after init(), since initially the Map is empty *)
let cterm s =
  try
    CTmap.find s (!terms)
  with Not_found ->
    raise (Internal_Error ("'" ^ s ^ "' not found in COQ terms"))

(* get the COQ term of a string. Full parsing (applications, functions, etc) *)
let term_of_string s =
  let constr_eoi = Pcoq.eoi_entry Pcoq.Constr.lconstr in
  let con_expr = Pcoq.parse_string constr_eoi s in
    Constrintern.interp_constr Evd.empty (Global.env ()) con_expr

(* Ocaml list to Coq list.
   The first argument is the type of the resulting COQ list *)
let list_to_clist t mll =
  let conv ml_e cl = Term.mkApp (cterm "cons", [| t; ml_e; cl |]) in
  List.fold_right conv mll (Term.mkApp (cterm "nil", [| t |]))

let int_to_N i =
  let str = string_of_int i in
  if i < 0 then
    raise (Internal_Error ("int_to_N " ^ str))
  else
    term_of_string (str^"%N")

let int_to_Z i =
  term_of_string ("("^(string_of_int i)^")%Z")

let lt_prop_t a b =
  Term.mkApp (cterm "N_lt_prop", [| a; b |])

let term_normalize env t =
  Closure.norm_val
    (Closure.create_clos_infos Closure.betadeltaiota env)
    (Closure.inject t)

let string_of_name n =
  match n with
    | Names.Name id -> Names.string_of_id id
    | Names.Anonymous -> "_"

let id_of_name n =
  match n with
  | Names.Name id -> id
  | Names.Anonymous -> raise (Internal_Error "id_of_name of Anonymous")

let tctx_push decl ectx tctx =
  let e_rel_ctx = Environ.rel_context ectx in
  let lift_decl d =
    let id, body, t = d in
    (id, body, Term.lift 1 t) in
  let e_rel_ctx' = List.map lift_decl e_rel_ctx in
  let ectx' = Environ.push_rel_context e_rel_ctx' Environ.empty_env in
  let tctx' = Environ.push_rel decl tctx in
  (ectx', tctx')

let term_type ctx t =
  Environ.j_type (Typeops.typing ctx t)

let term_equal_env env t1 t2 =
  Term.eq_constr (term_normalize env t1) (term_normalize env t2)

let whnf_env ctx t =
  Reduction.whd_betadeltaiota ctx t

(* wrappers for type Omega *)
let omega ()    = cterm "Omega"
let snat t      = Term.mkApp (cterm "snat", [| t |])
let sint t      = Term.mkApp (cterm "sint", [| t |])
let sbool t     = Term.mkApp (cterm "sbool", [| t |])
let arrow t1 t2 = Term.mkApp (cterm "arrow", [| t1; t2 |])
let tuple tl =
  let len = int_to_N (List.length tl) in
  let clist = list_to_clist (omega ()) tl in
  let list_to_lambda = cterm "list_to_lambda" in
  let e0 = snat (int_to_N 42) in
  Term.mkApp (cterm "tuple",
    [| len; Term.mkApp (list_to_lambda, [| (omega ()); clist; e0 |]) |] )
let forallt t f = Term.mkApp (cterm "forallT", [| t; f |])
let existst t f = Term.mkApp (cterm "existsT", [| t; f |])

(* wrappers for type OmegaK *)
let omegaK ()    = cterm "OmegaK"
let snatK t      = Term.mkApp (cterm "snatK", [| t |])
let sintK t      = Term.mkApp (cterm "sintK", [| t |])
let sboolK t     = Term.mkApp (cterm "sboolK", [| t |])
let arrowK t     = Term.mkApp (cterm "arrowK", [| t |])
let tupleK tl =
  let len = int_to_N (List.length tl) in
  let clist = list_to_clist (omegaK ()) tl in
  let list_to_lambda = cterm "list_to_lambda" in
  let e0 = snatK (int_to_N 42) in
  Term.mkApp (cterm "tupleK",
    [| len; Term.mkApp (list_to_lambda, [| (omegaK ()); clist; e0 |]) |] )
let foralltK t f = Term.mkApp (cterm "forallTK", [| t; f |])
let existstK t f = Term.mkApp (cterm "existsTK", [| t; f |])
let funcK t      = Term.mkApp (cterm "funcK", [| t |])

let to_omegaK t = Term.mkApp (cterm "to_OmegaK", [| t |])


(* Generic destructor for inductive types *)

let dest_inductive ctx indt t =
  try
    let (const, args) = Term.destApp (whnf_env ctx t) in
    let (tt, index) = Term.destConstruct const in
    if Term.eq_constr (Term.mkInd tt) indt then
      (index, args)
    else
      raise (Bad_type t)
  with Invalid_argument _ -> raise (Bad_type t)


(* Omega destructors *)

let dest_arithmetic ctx t =
  match dest_inductive ctx (omega ()) t with
  | 1, [| arg |] -> arg, false
  | 2, [| arg |] -> arg, true
  | _ -> raise (Bad_type t)

let dest_bool ctx t =
  match dest_inductive ctx (omega ()) t with
  | 3, [| arg |] -> arg
  | _ -> raise (Bad_type t)

let dest_arrow ctx t =
  match dest_inductive ctx (omega ()) t with
  | 4, [| arg0; arg1 |] -> (arg0, arg1)
  | _ -> raise (Bad_type t)

let dest_tuple ctx t =
  match dest_inductive ctx (omega ()) t with
  | 5, [| arg0; arg1 |] -> (arg0, arg1)
  | _ -> raise (Bad_type t)

let dest_forall ctx t =
  match dest_inductive ctx (omega ()) t with
  | 6, [| arg0; arg1 |] -> (arg0, arg1)
  | _ -> raise (Bad_type t)

let dest_exists ctx t =
  match dest_inductive ctx (omega ()) t with
  | 7, [| arg0; arg1 |] -> (arg0, arg1)
  | _ -> raise (Bad_type t)


(* OmegaK destructors *)

let dest_arithmeticK ctx t =
  match dest_inductive ctx (omegaK ()) t with
  | 1, [| arg |] -> arg, false
  | 2, [| arg |] -> arg, true
  | _ -> raise (Bad_type t)

let dest_boolK ctx t =
  match dest_inductive ctx (omegaK ()) t with
  | 3, [| arg |] -> arg
  | _ -> raise (Bad_type t)

let dest_arrowK ctx t =
  match dest_inductive ctx (omegaK ()) t with
  | 4, [| arg |] -> arg
  | _ -> raise (Bad_type t)

let dest_tupleK ctx t =
  match dest_inductive ctx (omegaK ()) t with
  | 5, [| arg0; arg1 |] -> (arg0, arg1)
  | _ -> raise (Bad_type t)

let dest_forallK ctx t =
  match dest_inductive ctx (omegaK ()) t with
  | 6, [| arg0; arg1 |] -> (arg0, arg1)
  | _ -> raise (Bad_type t)

let dest_existsK ctx t =
  match dest_inductive ctx (omegaK ()) t with
  | 7, [| arg0; arg1 |] -> (arg0, arg1)
  | _ -> raise (Bad_type t)

let dest_funcK ctx t =
  match dest_inductive ctx (omegaK ()) t with
  | 8, [| arg |] -> arg
  | _ -> raise (Bad_type t)


let cps_correctness t = Term.mkApp (cterm "cps_correctness", [| t |])

let init () =
  let add_term s =
    terms := CTmap.add s (term_of_string ("@"^s)) (!terms) in
  List.iter add_term coq_terms

let binop_N op t1 t2 =
  match op with
  | Clops.Barithop Clops.BOplus -> Term.mkApp (cterm "N_plus", [| t1; t2 |])
  | Clops.Barithop Clops.BOminus -> Term.mkApp (cterm "N_minus", [| t1; t2 |])
  | Clops.Barithop Clops.BOtimes -> Term.mkApp (cterm "N_times", [| t1; t2 |])
  | Clops.Barithop Clops.BOdiv -> Term.mkApp (cterm "N_div", [| t1; t2 |])
  | Clops.Barithop Clops.BOmod -> Term.mkApp (cterm "N_mod", [| t1; t2 |])
  | Clops.Brelop Clops.BOeq -> Term.mkApp (cterm "N_eq", [| t1; t2 |])
  | Clops.Brelop Clops.BOne -> Term.mkApp (cterm "N_ne", [| t1; t2 |])
  | Clops.Brelop Clops.BOlt -> Term.mkApp (cterm "N_lt", [| t1; t2 |])
  | Clops.Brelop Clops.BOle -> Term.mkApp (cterm "N_le", [| t1; t2 |])
  | Clops.Brelop Clops.BOgt -> Term.mkApp (cterm "N_gt", [| t1; t2 |])
  | Clops.Brelop Clops.BOge -> Term.mkApp (cterm "N_ge", [| t1; t2 |])
  | Clops.Blogicop _ ->
    raise (Internal_Error "binop_N with logical operator!")

let binop_Z op t1 t2 =
  match op with
  | Clops.Barithop Clops.BOplus -> Term.mkApp (cterm "Z_plus", [| t1; t2 |])
  | Clops.Barithop Clops.BOminus -> Term.mkApp (cterm "Z_minus", [| t1; t2 |])
  | Clops.Barithop Clops.BOtimes -> Term.mkApp (cterm "Z_times", [| t1; t2 |])
  | Clops.Barithop Clops.BOdiv -> Term.mkApp (cterm "Z_div", [| t1; t2 |])
  | Clops.Barithop Clops.BOmod -> Term.mkApp (cterm "Z_mod", [| t1; t2 |])
  | Clops.Brelop Clops.BOeq -> Term.mkApp (cterm "Z_eq", [| t1; t2 |])
  | Clops.Brelop Clops.BOne -> Term.mkApp (cterm "Z_ne", [| t1; t2 |])
  | Clops.Brelop Clops.BOlt -> Term.mkApp (cterm "Z_lt", [| t1; t2 |])
  | Clops.Brelop Clops.BOle -> Term.mkApp (cterm "Z_le", [| t1; t2 |])
  | Clops.Brelop Clops.BOgt -> Term.mkApp (cterm "Z_gt", [| t1; t2 |])
  | Clops.Brelop Clops.BOge -> Term.mkApp (cterm "Z_ge", [| t1; t2 |])
  | Clops.Blogicop _ ->
    raise (Internal_Error "binop_Z with logical operator!")

let binop_bool op t1 t2 =
  match op with
  | Clops.Blogicop Clops.BOand -> Term.mkApp (cterm "bool_and", [| t1; t2 |])
  | Clops.Blogicop Clops.BOor -> Term.mkApp (cterm "bool_or", [| t1; t2 |])
  | Clops.Barithop _ ->
    raise (Internal_Error "binop_bool with arithmetic operator!")
  | Clops.Brelop _ ->
    raise (Internal_Error "binop_bool with relative operator!")

let unop_bool op t =
  match op with
  | Clops.UOnot -> Term.mkApp (cterm "bool_not", [| t |])

let rec eq_OmegaK tctx t1 t2 =
(*
         Pp.ppnl_with Format.std_formatter (
            v 0 (str "CHECK" ++ brk (0,0) ++
              hv 0 (
                hov 0 (Printer.pr_constr_env tctx (whnf_env tctx t1)) ++ brk (1,0) ++
                str "==" ++ brk (1, 0) ++
                hov 0 (Printer.pr_constr_env tctx (whnf_env tctx t2)) ++
                str "--" ++ brk (1, 0)
                )));
         *)

  if term_equal_env tctx t1 t2 then
    true
  else
  try
    let (t1const, t1args) = Term.destApp (whnf_env tctx t1) in
    let (t2const, t2args) = Term.destApp (whnf_env tctx t2) in
    let (t1_typ, t1cind) = Term.destConstruct t1const in
    let (t2_typ, t2cind) = Term.destConstruct t2const in
    if term_equal_env tctx (Term.mkInd t1_typ) (cterm "OmegaK") &&
       term_equal_env tctx (Term.mkInd t2_typ) (cterm "OmegaK") then
      match t1cind, t2cind with
      | 1, 1 (* snatK *)
      | 2, 2 (* sintK *)
      | 3, 3 -> (* sboolK *)
        term_equal_env tctx (Array.get t1args 0) (Array.get t2args 0)
      | 4, 4 -> (* arrowK *)
        eq_OmegaK tctx (Array.get t1args 0) (Array.get t2args 0)
      | 5, 5 -> (* tupleK *)
        begin
        match t1args, t2args with
        | [| t1len; t1map |], [| t2len; t2map |] ->
          if term_equal_env tctx t1len t2len then
            let rec check_elem ind =
              if term_equal_env tctx (int_to_N ind) t1len then
                true
              else
                let t1_el = Term.mkApp (t1map, [| int_to_N ind |]) in
                let t2_el = Term.mkApp (t2map, [| int_to_N ind |]) in
                if eq_OmegaK tctx t1_el t2_el then
                  check_elem (ind + 1)
                else
                  false in
            check_elem 0
          else
            false
        | _ ->
          false
        end
      | 6, 6 (* forallTK *)
      | 7, 7 (* existsTK *) ->
        begin
        match t1args, t2args with
        | [| t1typ; t1map; |], [| t2typ; t2map; |] ->
          if term_equal_env tctx t1typ t2typ then
            let eempty = Environ.empty_env in
            let newvar = Names.Name (Names.id_of_string "__new") in
            let tctx' = snd (tctx_push (newvar, None, t1typ) eempty tctx) in
            let t1_prod = Term.mkApp (t1map, [| Term.mkRel 1 |]) in
            let t2_prod = Term.mkApp (t2map, [| Term.mkRel 1 |]) in
            eq_OmegaK tctx' t1_prod t2_prod
          else
            false
        | _ ->
          false
        end
      | 8, 8 -> (* funcK *)
        eq_OmegaK tctx (Array.get t1args 0) (Array.get t2args 0)
      | _ -> false
    else
      raise (Internal_Error "eq_OmegaK called with non-OmegaK terms!")
  with Invalid_argument _ -> false

let to_closure t1 t2 = Term.mkApp (cterm "to_closure", [| t1; t2 |])

let pr_ctx tctx ectx =
  let rec pr_tctx tctx tctx_rel =
    match tctx_rel with
    | [] -> (mt ())
    | [b] -> str "(" ++ Printer.pr_rel_decl tctx b ++ str ")"
    | b::rest ->
        let pb = Printer.pr_rel_decl tctx b in
        let penvtl = pr_tctx (Environ.push_rel b tctx) rest in
        str "(" ++ pb ++ str ")" ++ spc () ++ penvtl in
  let rec pr_ectx ectx_rel =
    match ectx_rel with
    | [] -> (mt ())
    | [b] -> str "(" ++ Printer.pr_rel_decl tctx b ++ str ")"
    | b::rest ->
        let pb = Printer.pr_rel_decl tctx b in
        let penvtl = pr_ectx rest in
        str "(" ++ pb ++ str ")" ++ spc () ++ penvtl in
  hov 0 (pr_tctx tctx (List.rev (Environ.rel_context tctx))) ++
  brk(1, 0) ++
  hov 0 (pr_ectx (List.rev (Environ.rel_context ectx)))
