
type term = Term.constr

exception Bad_type of term
exception Internal_Error of string
exception Type_Error of Pp.std_ppcmds

(* string to COQ constructor *)
val cterm : string -> term

(* Omega type *)
val snat    : term -> term
val sint    : term -> term
val sbool   : term -> term
val arrow   : term -> term -> term
val tuple   : term list -> term
val forallt : term -> term -> term
val existst : term -> term -> term

val omega   : unit -> term

(* OmegaK type *)
val snatK    : term -> term
val sintK    : term -> term
val sboolK   : term -> term
val arrowK   : term -> term
val tupleK   : term list -> term
val foralltK : term -> term -> term
val existstK : term -> term -> term
val funcK    : term -> term

val omegaK : unit -> term

(* CPS transformation. Omega -> OmegaK *)
val to_omegaK : term -> term

(* CPS correctness theorem, Omega -> OmegaK *)
val cps_correctness : term -> term

(* closure transformation. OmegaK -> OmegaK -> OmegaK *)
val to_closure : term -> term -> term

(* get the COQ term of a string. Full parsing (applications, functions, etc) *)
(* val term_of_string : string -> term *)

(* get a COQ nat (N) or int (Z) from an int *)
val int_to_N : int -> term
val int_to_Z : int -> term

(* "lt_prop_X a b" produces a proof that a < b *)
val lt_prop_t : term -> term -> term

(* binary operations *)
val binop_N : Clops.binop -> term -> term -> term
val binop_Z : Clops.binop -> term -> term -> term
val binop_bool : Clops.binop -> term -> term -> term

(* unary operators *)
val unop_bool : Clops.unop -> term -> term

(* Get the string of a COQ name (variable) *)
val string_of_name : Names.name -> string
(* Get the identifier of a COQ name (variable) *)
val id_of_name : Names.name -> Names.identifier

(* push a term in the typing context.
   first env is expression env, second is typing env
   result is the extended environments : (expression env * typing env) *)
val tctx_push :
  Term.rel_declaration -> Environ.env -> Environ.env ->
  (Environ.env * Environ.env)

(* normalize term (beta, delta, iota reductions) *)
val term_normalize : Environ.env -> term -> term

(* term Typing *)
val term_type : Environ.env -> term -> term

(* Check the equality of two terms. Beta, Delta, Iota, Zeta convertion *)
val term_equal_env : Environ.env -> term -> term -> bool

(* weak normal form (beta delta iota reduction) *)
val whnf_env : Environ.env -> term -> term

(*
   Generic destructor for an inductive term.

   If t1 is an (closed) inductive type with constructors C1, C2, ...
   and t2 is a term "Cn arg0 ... argn"
   then "dest_inductive ctx t1 t2" returns:
   (n, [| arg0; ....; argn |]

   If t2 does not belong in t1, Bad_type is raised.
   Partial applications also raise an exception

   Constructor indexes start from 1
*)
val dest_inductive :
  Environ.env -> term -> term -> int * term array

(* destructors for the Omega type *)

(* (term (:Z), true) if it is sint
   (term (:N), false) if it is snat *)
val dest_arithmetic : Environ.env -> term -> term * bool
val dest_bool : Environ.env -> term -> term
val dest_arrow : Environ.env -> term -> term * term
val dest_tuple : Environ.env -> term -> term * term
val dest_forall : Environ.env -> term -> term * term
val dest_exists : Environ.env -> term -> term * term

(* destructors for the OmegaK type *)

(* (term (:Z), true) if it is sintK
   (term (:N), false) if it is snatK *)
val dest_arithmeticK : Environ.env -> term -> term * bool
val dest_boolK : Environ.env -> term -> term
val dest_arrowK : Environ.env -> term -> term
val dest_tupleK : Environ.env -> term -> term * term
val dest_forallK : Environ.env -> term -> term * term
val dest_existsK : Environ.env -> term -> term * term
val dest_funcK : Environ.env -> term -> term

(* equality toy OmegaK terms *)
val eq_OmegaK : Environ.env -> term -> term -> bool

val init : unit -> unit

val pr_ctx : Environ.env -> Environ.env -> Pp.std_ppcmds
